/* Implementation of Edinburgh Compatible Context Editor
   in LC for Intel 8086 by Rainer Thonnes, December 1980
   based on that in IMP for DEC PDP9/15 by Hamish Dewar,
   June 1970 */

// Only one known current bug:  typing the 'y' macro command before it is
// defined causes the program to hang.  Probably in an infinite loop.

var min=1
var sin=2
var stop=-5000
var in=min
var mon=0
var casebit=95
var print1=0
var print2=0
var sextra=1122
var pp1
var sym
var code
var text
var num
var cbase=1
var tbase=120
var ci
var ti
var cmax=0
array c[121]
array stored[193]
var pos1=0
var pos2=0
var pos3=0
var bufmax=510000
var top=1
var bot=509878
array a[510001]
var lbeg
var pp
var fp=0
var lend=0
var fend=0
var ms=0
var ml=0
var type
var chain
var pend=0
var mfp=0
var mlend=0
var mend=0
var sfp=0
var send=0
const array symtype[96] =
   0   0   0   0   0   0   0   0   0   0   0   0   0   0   0   0
   0   0   0   0   0   0   0   0   0   0   0   0   0   0   0   0

   0  64   3   3   3   2   3   3  11   9  64   3  12   2   3   3
   0   0   0   0   0   0   0   0   0   0   3   1   3   3   3  64
   3  18  10  26   5   8  52  10   2   6  10  10 122  56   2   2
  10  50  10  22   5 117   6   2  32  32  32   3  10   3   3   3;

/*    !   "   #   $   %   &   '   (   )   *   +   ,   -   .   /
  0   1   2   3   4   5   6   7   8   9   :   ;   <   =   >   ?
  @   A   B   C   D   E   F   G   H   I   J   K   L   M   N   O
  P   Q   R   S   T   U   V   W   X   Y   Z   [   \   ]   ^   _

Special classes

0 Normal
1 %A,%C,%S
2 %X,%Y,%Z
3 %M,%F,%Q
4 ! * ?
5 unused
6 unused
7 %L,%U

Normal classes

0 Digits 0:9
1 NL and ;
2 Illegals %,-,A,H,N,O,Q,W
3 delimiters "#$&'+./<>=@[]^_
4 F
5 D,T,U
6 I,S,V
7 unused
8 E,M
9 )
10 B,C,G,J,K,L,P,R,\
11 (
12 ,
13 unused
14 unused
15 unused */

var i
var j
var k
var nl=10

proc mainfp()
  return fp if in=min
  return mfp
end

proc readsym()
  if pend<>0
    sym=pend; pend=0
  else
    cycle
      if pos3=0
        sym=consoleget(); return
      finish
      sym=stored[pos3]; pos3=pos3+1
      return if sym<>0
      pos3=pos2; pos2=pos1; pos1=0
    repeat
  finish
end

proc readitem()
  cycle
    type=1
    cycle
      readsym(); return -1 if sym < 0;
    repeat if sym=' '
    return 0 if sym<' '
    sym=sym-32 if sym>=96
    type=symtype[sym]
    return 0 if type&15<>0
    if type=32
      pos1=pos2; pos2=pos3; pos3=((sym-'X')<<6)+1
    finish
  repeat if type=32
  if type=0
    num=sym-'0'
    cycle
      readsym(); return 0 if sym < 0; pend=sym
      return 0 if sym<'0'
      return 0 if sym>'9'
      pend=0; num=num*10-'0'+sym
    repeat
  else
    type=0; num=0; return 0 if sym='*'
    num=stop+1;    return 0 if sym='?'
    num=stop                /* sym='!' */
  finish
  return 0
end

proc unchain()
  cycle
    text=chain; return if text=0
    chain=c[text+1]; c[text+1]=ci
  repeat if c[text]<>'('
end

proc stack(v)
  c[ci]=v; ci=ci+1
end

proc readline()
var k
  if fp<>fend
    lend=fp
    cycle
      return if a[lend]=nl
      lend=lend+1
    repeat
  finish
  fp=bot-1121
  cycle
    k=nl;
    k=fileget(in) if fp<>bot;
    if k < 0
      fp=bot; lend=fp; fend=lend; a[fp]=nl
      jump ret
    finish
    a[fp]=k; fp=fp+1
  repeat if k<>nl
  fend=fp; lend=fend-1; fp=bot-1121
ret: ms=0; print1=0; print2=0
end

proc breakline()
  a[pp]=nl; pp=pp+1; lbeg=pp
end

proc leftstar()
  cycle
    return if pp=lbeg
    fp=fp-1; pp=pp-1; a[fp]=a[pp]
  repeat
end

proc rightstar()
  cycle
    return if fp=lend
    a[pp]=a[fp]; pp=pp+1; fp=fp+1
  repeat
end

proc makespace()
var k
var p1
var p2
  return if mainfp()-pp>240
  p1=top; p2=(p1+lbeg)>>1; p2=lbeg if code='C'
  dosignal(15, p1, p2) if p2=top
  cycle
    k=a[p1]; putsym(k); p1=p1+1
    k=0 if p1<p2
  repeat if k<>nl
  lbeg=top+lbeg-p1; p2=pp; pp=top
  cycle
    return if p1=p2
    a[pp]=a[p1]; pp=pp+1; p1=p1+1
  repeat
end

proc refresh()
  fp=fp+1; makespace(); readline()
end

proc printline()
var p
  print1=lend; print2=fp+pp; p=lbeg
  cycle
    if p=pp
      if p<>lbeg; if num=0; psym('^'); finish; finish
      p=fp
    finish
    if p=lend
      if p=fend
        psym('*'); psym('*'); psym('E'); psym('n'); psym('d'); psym('*'); psym('*')
      finish
      psym(nl); return
    finish
    psym(a[p]); p=p+1
  repeat
end

proc cased(ch)
  return ch if ch<'a'
  return ch if ch>'z'
  return ch&casebit
end

proc matched()
var i
var j
var k
var l
var t1
var fp1
var lim
  lim=c[ci-3]>>7
  t1=cased(c[text])
  cycle
    pp1=pp; fp1=fp;
    if fp<>lend; cycle
      k=a[fp];
      if cased(k)=t1
        if fp=ms
          jump no if code='F'
          jump no if code='U'
        finish
        i=fp; j=text
        cycle
          i=i+1; j=j-1; l=cased(c[j]);
          if l=0
            ms=fp; ml=i; return 1
          finish
        repeat if cased(a[i])=l
      finish
      return 0 if code='V'
no:   a[pp]=k; pp=pp+1; fp=fp+1
    repeat if fp<>lend; finish
    return 0 if code='V'
    lim=lim-1; lim=0 if fp=fend
    if lim=0
      pp=pp1; fp=fp1; return 0
    finish
    if code='U'
      pp=pp1
    else
      breakline()
    finish
    refresh()
  repeat
end

proc switchinputs()
  if in=min
    leftstar(); in=sin
    mfp=fp; mlend=lend; mend=fend
    bot=bot+sextra; fp=sfp; fend=send
    readline()
  else
    pp=lbeg; in=min
    bot=bot-sextra; sfp=fp; send=fend; fp=mfp
    lend=mlend; fend=mend
  finish
end

  psym('E'); psym('C'); psym('C'); psym('E')
  psym(' '); psym('0'); psym('6'); psym('/'); psym('0')
  psym('1'); psym('/'); psym('8'); psym('1'); psym(nl)

  pp=top-1; breakline(); readline()
readco: fflush(0); xprompt('>'); xprompt('>') if in=sin
  cycle
    ci=cbase; ti=tbase; chain=0
    jump eof if readitem()<0;
  repeat if type=1
  if type=0; if cmax<>0
    c[cmax+2]=num; jump eof if readitem()<0;
    jump er2 if type<>1
    jump execute
  finish; finish
  if sym='%'
    readsym(); jump eof if sym < 0; sym=sym-32 if sym>=96; code=sym
    jump er5 if code<'A'; jump eof if readitem()<0;
    i=symtype[code]>>4
  else
nextunit: i=type&15; jump er2 if i<4
    code=sym; text=0; num=1; jump eof if readitem()<0;
  finish
  jump er5 if i=0
  if i=1
    if code='A'
      dosignal(0, 0, 0)
    finish
    if code='S'
      switchinputs(); jump monitor
    finish
eof:code='C'
    switchinputs() if in=sin
    cycle
      rightstar();
      if fp=fend
        cycle
          dosignal(0, 0, 0) if top=pp
          putsym(a[top]); top=top+1
        repeat
      finish
      breakline(); refresh()
    repeat
  finish
  if i=2
    i=((code-'X')<<6)+1;
    if sym='='
      cycle
        readsym(); jump eof if sym < 0;
        if sym<>nl
          stored[i]=sym
          jump er6 if i&63=0
          i=i+1
        finish
      repeat if sym<>nl
      stored[i]=0
      jump readco
    finish
    cycle
      sym=stored[i];
      if sym=0
        psym(nl); jump readco
      finish
      psym(sym); i=i+1
    repeat
  finish
  if i=3
    mon='M'-code; jump readco
  finish
  if i=4
    num=0 if type<>0
  finish
  if i<=5
    code=(num<<7)+code; num=1;
    if type=0
      jump eof if readitem()<0;
    finish
  finish
  if i<=6
    jump er4 if type<>3; text=ti; i=sym
    cycle
      readsym(); jump eof if sym < 0;
      if sym=nl
        pend=sym; i=sym
      finish
      if sym=i
        if code<>'I'; if code<>'S'
          jump er4 if sym=nl
          jump er4 if ti=text
        finish; finish
        c[ti]=0; ti=ti-1; jump ri
      finish
      jump er6 if ti<=ci; c[ti]=sym; ti=ti-1
    repeat
  finish
  if i=7
    casebit=95; casebit=127 if code='L'; jump readco
  finish
  if i=8
    jump nq if sym<>'-'; code=code+10
ri: jump eof if readitem()<0; jump rn
  finish
  if i=9
    unchain(); jump er3 if text=0
    c[text+2]=num; text=text+3
  finish
  if i<=10
nq: jump er1 if type=3
rn: if type=0
      jump eof if readitem()<0;
    finish
    jump put
  finish
  if i=12
    if type=1
      jump eof if readitem()<0
    finish
  finish
  if i<=12
    text=chain; chain=ci; num=0
  finish
  dosignal(14, 1, i) if i>12
put: stack(code); stack(text); stack(num)
  jump er6 if ci+4>=ti
  jump nextunit if type<>1
  unchain(); jump er3 if text<>0
  cmax=ci
  stack(')'); stack(cbase); stack(1)
  stack(0); jump execute
er1: psym(' '); psym(code)
er2: code=sym; jump er5
er3: psym(' '); psym('('); psym(')'); jump er7
er4: psym(' '); psym('T'); psym('e'); psym('x'); psym('t')
     psym(' '); psym('f'); psym('o'); psym('r')
er5: psym(' '); psym(code&127); jump er7
er6: psym(' '); psym('S'); psym('i'); psym('z'); psym('e')
er7: psym('?'); psym(nl)
  cmax=0 if ci<>cbase
  cycle
    jump readco if sym=nl; readsym(); jump eof if sym < 0;
  repeat
execute:
  ci=cbase
get: code=c[ci]&127; jump monitor if code=0
  text=c[ci+1]; num=c[ci+2]; ci=ci+3
rep: num=num-1
  jump xm if code='M'
  jump xw if code='W'
  jump xl if code='L'
  jump xr if code='R'
  jump xc if code='C'
  jump xlb if code='('
  jump xcomma if code=','
  jump xrb if code=')'
  jump no if code='\\'
  jump xj if code='J'
  jump xk if code='K'
  jump xp if code='P'
  jump xb if code='B'
  jump xg if code='G'
  jump xi if code='I'
  jump xs if code='S'
  jump xt if code='T'
  jump xd if code='D'
  jump xu if code='U'
  jump xf if code='F'
  jump xv if code='V'
  jump xo if code='O'
  jump xe if code='E'
  dosignal(14, 2, code)
ok: jump get if num=0
    jump get if num=stop
    jump rep
no: jump get if num<0
  if c[ci]='\\'
    ci=ci+3; jump get
  finish
skp: i=c[ci]; ci=c[ci+1] if i='('; ci=ci+3;
  if i<>','; if i<>')'
    jump skp if i<>0; jump xer
  finish; finish
  num=c[ci-1]-1; jump no
xer: psym('F'); psym('a'); psym('i'); psym('l'); psym('e'); psym('d')
  psym(' ');
  if code='O'
    psym('E'); code='-'
  else if code='W'
    psym('M'); code='-'
  finish
  psym(code);
  if text<>0
    psym('"')
    cycle
      i=c[text]; psym(i) if i<>0; text=text-1
    repeat if i<>0
    psym('"')
  finish
  psym(nl); print1=0
monitor:
  jump readco if sym<>nl
  jump readco if mon<0
  if print1=lend
    jump readco if mon=0
    jump readco if print2=fp+pp
  finish
  num=0; printline()
  jump readco
xlb: c[text+2]=num+1; jump get
xrb: jump get if num=0
     jump get if num=stop
     c[ci-1]=num
xcomma: ci=text; jump get
xc: jump no if fp=lend; i=a[fp];
    if i&95>='A'; if i&95<='Z'
      if i&32 <> 0
        a[fp]=i-32
      else
        a[fp]=i+32
      finish
    finish; finish
xr: jump no if fp=lend
    a[pp]=a[fp]; pp=pp+1; fp=fp+1; jump ok
xl: jump no if pp=lbeg
    jump no if in=sin
    fp=fp-1; pp=pp-1; a[fp]=a[pp]
    ms=0; jump ok
xe: jump no if fp=lend; fp=fp+1; jump ok
xo: jump no if pp=lbeg; pp=pp-1; jump ok
xv:xu:xd:xf:    jump no if matched()=0
    if code='U'
      pp=pp1
    else if code='D'
      fp=ml
    finish
    jump ok
xt: jump no if matched()=0
    cycle
      a[pp]=a[fp]; pp=pp+1; fp=fp+1
    repeat if fp<>ml
    jump ok
xs: jump no if fp<>ms; fp=ml
xi: makespace()
    jump no if pp-lbeg+lend-fp>80
  i=text
  cycle
    jump ok if c[i]=0
    a[pp]=c[i]; pp=pp+1; i=i-1
  repeat
xg: fflush(0); xprompt(':')
    makespace(); i=consoleget(); jump eof if i<0; jump no if i=':'; leftstar();
  if i<>nl; cycle
    a[pp]=i; pp=pp+1; i=consoleget(); jump eof if i<0;
  repeat if i<>nl; finish
xb: breakline(); jump ok
xp: printline(); jump get if num=0
xm: rightstar(); jump no if fp=fend; breakline()
m1: refresh(); jump ok
xk: pp=lbeg; fp=lend
k1: jump no if fp=fend; jump m1
xj: rightstar(); jump no if pp-lbeg>80; jump k1
xw: jump no if in=sin
    makespace()
  jump no if lbeg=top
  lend=fp-pp+lbeg-1
  cycle
    k=a[pp-1];
    if k=nl; if pp<>lbeg
      lbeg=pp; ms=0; jump ok
    finish; finish
    fp=fp-1; pp=pp-1; a[fp]=k
  repeat
  dosignal(0, 0, 0)
