# BIPS (Built-in Phrases) are linked to the type-code returned
# by the line-reconstruction code (aka lexer)

# These *must* come first.

B<IDENT>=1;
B<NUM>=5;
B<CHARLIT>=3;

# Phrase definitions.  SSLIST is the main entry point.


P<SSLIST> = <STATEMENT> <SSLIST>,
            ;
P<SYNTAXERROR> = ;
P<EXISTINGVAR> = <IDENT>;
P<NEWVAR> = <IDENT>;

P<STATEMENT> = <EXISTINGVAR> ':' <SIMPLE> <OPTSEMI>,
               <SIMPLE> <OPTSEMI>;

P<SIMPLE> = <SS>,
            <CONSTDECL>,
            <VARDECL>,
            <DEFFN>,
            <ARRAY>,
            <EXTERN>,
            <PROCDEF>
            ;
P<EXTERN> = "extern" <EXISTINGVAR>;
P<PROCDEF> = "proc" <EXISTINGVAR> '(' <ARGLIST> ')' <OPTSEMI> <PROCBODY> "end";
P<PROCBODY> = <SS> <OPTSEMI> <PROCBODY>,
              ;
P<ARGLIST> = <EXISTINGVAR> <ARGLIST>,
             ;

P<CONSTDECL> = "const" <EXISTINGVAR> '=' <NUM>;

P<VARDECL> = "var" <VARDEC> <VARDECLIST>;

P<VARDEC> = <IDENT> '=' <NUM>,
            <IDENT>;

P<VARDECLIST> = ',' <VARDEC> <VARDECLIST>,
              ;

P<ARRAY> = "initialised" "array" <EXISTINGVAR> '[' <NUM> ']' '=' <INITLIST> <OPTSEMI>,
           "array" <IDENT> '[' <NUM> ']';

P<INITLIST> = <NUM> <INITLIST>,
              ;

P<DEFFN> = "let" <EXISTINGVAR> '(' <FORMALS> ')' '=' <BOOLEXPR>;

P<FORMALS> = <EXISTINGVAR> <RESTOFFORMALS>;

P<RESTOFFORMALS> = ',' <EXISTINGVAR> <RESTOFFORMALS>, ;

P<SS> = <EXISTINGVAR> '=' <BOOLEXPR> <OPTIF>,
        <EXISTINGVAR> '[' <EXPR> ']' '=' <BOOLEXPR> <OPTIF>,
        <PROCCALL> <OPTIF>,
        <IFSEQ>,
        <LOOP>,
        "return" <BOOLEXPR> <OPTIF>,
        "return" <OPTIF>,
        "jump" <EXISTINGVAR> <OPTIF>;
P<IFSEQ> = "if" <BOOLEXPR> <OPTSEMI> <THENPART> <OPTELSEPART> "finish";
P<THENPART> = <SS> <OPTSEMI> <THENPART>,
              ;
P<OPTELSEPART> = "else" <THENPART>,
               ;
P<ELSEIFLIT> = "else" "if", "elseif";
P<LOOP> = "cycle" <LOOPPART> "repeat" <OPTIF>;
P<LOOPPART> = <SS> <OPTSEMI> <LOOPPART>,
              ;

P<PROCCALL> = <EXISTINGVAR> '(' <PARAMLIST> ')';

P<OPTIF> = "if" <BOOLEXPR>,
           ;

P<BOOLEXPR> = <BOOLTERM> <RESTOFBOOLTERM>;

P<RESTOFBOOLTERM> = <OROP> <BOOLTERM> <RESTOFBOOLTERM>,
                    ;

P<BOOLTERM> = <BOOLFACTOR> <RESTOFBOOLFACTOR>;

P<RESTOFBOOLFACTOR> = <ANDOP> <BOOLFACTOR> <RESTOFBOOLFACTOR>,
                      ;

P<BOOLFACTOR> = <OPTNOT> <RELATION>;

P<OPTNOT> = '!',
            ;

P<OPTSEMI> = ';',
             ;

P<RELATION> = <EXPR> <RESTOFRELATION>;

P<RESTOFRELATION> = <RELOP> <EXPR>,
                    ;

P<EXPR> = <SUM> <RESTOFEXPR>;

P<RESTOFEXPR> = <SHIFTOP> <SUM> <RESTOFEXPR>,
                ;

P<SUM> = <OPTADDOP> <TERM> <RESTOFSUM>;

P<UNOP> = '+', '-', '\\';

P<ADDOP> = '+', '-';

P<OPTADDOP> = <ADDOP>,
              ;

P<RESTOFSUM> = <ADDOP> <TERM> <RESTOFSUM>,
               ;

P<TERM> = <FACTOR> <RESTOFTERM>;

P<RESTOFTERM> = <MULOP> <FACTOR> <RESTOFTERM>,
                ;

P<OROP> = '||';

P<ANDOP> = '&&';

P<RELOP> = '<>', '<=', '<', '>=', '>', '!=', '==', '#', '=';

P<SHIFTOP> = '<<', '>>';

P<MULOP> = '*', '/';

P<FACTOR> = '(' <BOOLEXPR> ')',
            <EXISTINGVAR> '(' <PARAMLIST> ')',
            <EXISTINGVAR> '[' <EXPR> ']',
            <EXISTINGVAR>,
            <NUMBER>;

 P<PARAMLIST> = <EXPR> <RESTOFPARAMLIST>,
              ;


 P<RESTOFPARAMLIST> = ',' <EXPR> <RESTOFPARAMLIST>,
               ;


P<NUMBER> = <NUM>,
            <UNOP> <NUM>,
            <CHARLIT>;

E
# 'E' is end of grammar.  Everything after this is ignored.
