#include <stdio.h>
#include <stdlib.h>


// the conclusion I've reached from writing up my thoughts for a compilers101 article
// is that the *usage* AST tree has to be built up from a combination of the concrete
// syntax tree (i.e. in the 'compile()' procedure) and the AST item already created for
// the object's definition.  the resulting object is an AST tree that closely parallels
// (and copies) the declaration but does not directly reference it.

//----------------------------------------------------------------------------------------

// The structure of symbol table that I'm looking for is fairly similar to that described here:
// http://compilers.iecc.com/comparch/article/99-01-062

// We have AST nodes that reflect the structure of a declaration.  All we need to add is
// a pointer in that AST node to the current lexical scope, and we keep an array of pointers
// to symbols described within each scope.  So to search for a symbol, we scan the most nested
// scope first, and work our way up to the top level.

// Naive implementation uses a stack of scopes, with each level being a linked list.  We can
// recode this for efficiency once the functional structure is worked out, and only then if
// it is absoluitely necessary.  Although a significant portion of compile time can be taken
// up by symbol table handling, with machines so fast nowadays, the difference between 'fucking fast'
// and just plain 'fast' is barely worth the effort in programmer time.  if it compiles in less
// than a second, I don't really care if I can make it twice as fast by using a hash table
// over a linked list.

// Looking up something like a struct field is done slightly differently than looking up a
// random variable or procedure.  Once you have looked up the top-level struct name, you have
// a pointer to its definition node, and finding the subfields is no longer a global search but
// is limited to the context of that declaration (i.e they are subfields of the AST for that struct)
// and indeed it would be wrong to look for a subfield by searching other scopes.  it's either a
// field in the struct being compiled or it is an error.

// The alternative to a stack of scopes, which links to strings, is to have a table of strings,
// each of which links to a stack of scopes.  This is more efficient when you have only a few
// cases where a tag is redefined at different scopes, and lends itself well to a hash table...

// I will probably end up trying both...

typedef int AST;

typedef struct symbol {
    char *name;       // this is what is being searched for in a particular context, eg a record field
    AST typeinfo;     // opaque index to generic object definition (eg type, size, member names etc)
    AST instanceinfo; // opaque index to specific object info (eg base reg, offset etc)
    struct symbol *nextsymbol; // next object via a linked list.  could equally be a has table
} symbol;

typedef struct symtab {
    char *name;     // name of this scope
    AST scopetype;  // procedure, function, unnamed block etc
    int scopedepth; // lexical nesting depth of this scope
    symbol *sym;    // all symbols in this scope
} symtab;

int current_scope = 0;
#define MAX_NESTING_DEPTH 10
symtab sym[MAX_NESTING_DEPTH];

// it's not just blocks. For example a procedure parameter list has its own
// scope for the parameter declarations which is different from both the
// enclosing block and the following procedure body

// whether we handle the lookup scope of usages for structured objects here
// too is TBD... (obj->fieldp.field[n]->i)

AST followtype(AST object, AST decl)
{
    // decorates the object with info from the decl, follows down both in parallel,
    // generates error if can't follow.

    // presumable doesn't follow index values as these will be handled separately?
    // ... or do we decorate the whole program tree in one massive pervasive tree walk?

    // or do we leave it all to the code generator to do on the fly?
    // which would require extra params for all codegen calls - *or* varargs!
    // (might solve short-circuit eval issue, but makes for less portability)
}

AST findtype(AST decl) // top-level entry point, only given 'obj' in example above
{
}

// so..?
void decorate(AST root, int scope) // int scope or AST scope?  can the symbol table be an AST tree too???
{
}


void push_scope(int AST, int type)
{
}

void pop_scope(int *AST)
{
}

int main(int argc, char **argv)
{
  fprintf(stderr, "Symbol table test\n");
  exit(0);
  return 0;
}
