# BIPS (Built-in Phrases) are linked to the type-code returned
# by the line-reconstruction code (aka lexer)

# These *must* come first.

B<IDENT>=1;
B<NUM>=5;
B<CHARLIT>=3;

# Phrase definitions.  PROGRAM is the main entry point.

P<PROGRAM> = <DECLARATIONS> <SSLIST>;
P<DECLARATIONS> = <VARDECL> <DECLARATIONS>,
                  <PROCDECL> <DECLARATIONS>,
                  <DEFFN> <DECLARATIONS>,
                  <ARRAY> <DECLARATIONS>,
                  <EXTERN> <DECLARATIONS>,
		       ;
P<PROCDECLARATIONS> = <VARDECL> <DECLARATIONS>,
                      <ARRAY> <DECLARATIONS>,
                      <EXTERN> <DECLARATIONS>,
		         ;
P<PROCDECL> = "proc" <EXISTINGVAR> '(' <ARGLIST> ')' <PROCDECLARATIONS> <SSLIST> "end" <OPTSEMI>;

P<SS> = <EXISTINGVAR> '=' <BOOLEXPR> <OPTIF> <OPTSEMI>,
        <EXISTINGVAR> '[' <EXPR> ']' '=' <BOOLEXPR> <OPTIF> <OPTSEMI>,
        <PROCCALL> <OPTIF> <OPTSEMI>,
        <IFSEQ> <OPTSEMI>,
        <LOOP> <OPTSEMI>,
        "while" '(' <BOOLEXPR> ')' "do" <SSLIST> "endwhile" <OPTSEMI>,
        "loop" <SSLIST> "endloop" <OPTSEMI>,
        "return" <BOOLEXPR> <OPTIF> <OPTSEMI>,
        "return" <OPTIF> <OPTSEMI>,
        "jump" <EXISTINGVAR> <OPTIF> <OPTSEMI>
      ;
P<SSLIST> = <OPTLABELS> <SS> <SSLIST>,
          ;

P<CONSTDECL> = "const" <EXISTINGVAR> '=' <NUMBER> <OPTSEMI>;

P<VARDECL> = "var" <VARDEC> <VARDECLIST> <OPTSEMI>;

P<VARDEC> = <IDENT> '=' <BOOLEXPR>,
            <IDENT>;
P<OPTLABELS> = <LABEL> <OPTLABELS>,
             ;
P<LABEL> = <EXISTINGVAR> ':';
P<SYNTAXERROR> = ;
P<EXISTINGVAR> = <IDENT>;
P<NEWVAR> = <IDENT>;
P<EXTERN> = "extern" <EXISTINGVAR> <OPTSEMI>;
P<ARGLIST> = <EXISTINGVAR> <ARGLIST>,
             ;

P<VARDECLIST> = ',' <VARDEC> <VARDECLIST>,
              ;

P<ARRAY> = "const" "array" <EXISTINGVAR> '[' <NUM> ']' '=' <INITLIST> <OPTSEMI>,
           "array" <IDENT> '[' <NUM> ']' <OPTSEMI>;

P<INITLIST> = <NUM> <INITLIST>,
              ;

P<DEFFN> = "let" <EXISTINGVAR> '(' <FORMALS> ')' '=' <BOOLEXPR> <OPTSEMI>;

P<FORMALS> = <EXISTINGVAR> <RESTOFFORMALS>;

P<RESTOFFORMALS> = ',' <EXISTINGVAR> <RESTOFFORMALS>, ;
P<IFSEQ> = "if" <BOOLEXPR> <OPTSEMI> <SSLIST> <OPTELSEPART> "finish";
P<OPTELSEPART> = "else" "if" <BOOLEXPR> <OPTSEMI> <SSLIST> <OPTELSEPART>,
                 "elseif" <BOOLEXPR> <OPTSEMI> <SSLIST> <OPTELSEPART>,
                 "else" <OPTSEMI> <SSLIST>,
               ;
P<LOOP> = "cycle" <OPTSEMI> <SSLIST> "repeat" <OPTIF>;

P<PROCCALL> = <EXISTINGVAR> '(' <PARAMLIST> ')';

P<OPTIF> = "if" <BOOLEXPR>,
           ;

P<BOOLEXPR> = <BOOLTERM> <RESTOFBOOLTERM>;

P<RESTOFBOOLTERM> = <BOROP> <BOOLTERM> <RESTOFBOOLTERM>,
                    ;

P<BOOLTERM> = <BOOLFACTOR> <RESTOFBOOLFACTOR>;

P<RESTOFBOOLFACTOR> = <BANDOP> <BOOLFACTOR> <RESTOFBOOLFACTOR>,
                      ;

P<BOOLFACTOR> = <OPTNOT> <RELATION>;

P<OPTNOT> = '!',
            ;

P<OPTSEMI> = ';',
             ;

P<RELATION> = <EXPR> <RESTOFRELATION>;

P<RESTOFRELATION> = <RELOP> <EXPR>,
                    ;

P<EXPR> = <SUM> <RESTOFEXPR>;

P<RESTOFEXPR> = <SHIFTOP> <SUM> <RESTOFEXPR>,
                ;

P<SUM> = <OPTUNOP> <TERM> <RESTOFSUM>;

P<UNOP> = '+', '-', '\\';

P<OPTUNOP> = <UNOP>,
              ;

P<ADDOP> = '+', '-';

P<OPTADDOP> = <ADDOP>,
              ;

P<RESTOFSUM> = <ADDOP> <TERM> <RESTOFSUM>,
               ;

P<TERM> = <BITFACTOR> <RESTOFTERM>;

P<RESTOFTERM> = <MULOP> <BITFACTOR> <RESTOFTERM>,
                ;

P<BITFACTOR> = <FACTOR> <RESTOFBITFACTOR>;

P<RESTOFBITFACTOR> = <BITOP> <FACTOR> <RESTOFBITFACTOR>,
                ;

P<BITOP> = '&';

P<BOROP> = '||';

P<BANDOP> = '&&';

P<RELOP> = '<>', '<=', '<', '>=', '>', '!=', '==', '#', '=';

P<SHIFTOP> = '<<', '>>';

P<MULOP> = '*', '/';

P<FACTOR> = '(' <BOOLEXPR> ')',
            <EXISTINGVAR> '(' <PARAMLIST> ')',
            <EXISTINGVAR> '[' <EXPR> ']',
            <EXISTINGVAR>,
            <NUMBER>;

 P<PARAMLIST> = <EXPR> <RESTOFPARAMLIST>,
              ;


 P<RESTOFPARAMLIST> = ',' <EXPR> <RESTOFPARAMLIST>,
               ;


P<NUMBER> = <NUM>,
            <UNOP> <NUM>,
            <CHARLIT>;

E
# 'E' is end of grammar.  Everything after this is ignored.
