// FILE IS BEING EDITED.  I'M ADDING FUNCTIONS RIGHT NOW.

// You are reading a C source file displayed using 'folds'.  Click on one of the
// sections below to unfold its contents.  Reclicking, or clicking on the expanded
// text, will collapse the expanded text back to the summary.

// If you are interested in the software used here (rather than the actual program,
// which is a demonstration compiler for the Yahoo 'compilers101' group) then
// you can read about it at my blog:
//
//       http://techennui.blogspot.com/2008/04/semi-literate-programming.html

/*< Include files, consts, incidental globals */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>

#ifndef FALSE
#define FALSE (0!=0)
#define TRUE (0==0)
#endif

#define DEBUG_TRIPS_AFTER TRUE
#define DEBUG_TREES TRUE

char *progname;

/*>*/

/*< Parser support */
#define DEBUG_PARSER
#include "varcalc.h" // GENERATED GRAMMAR FROM takeon.c and varcalc.g

int debug_parser = FALSE;

// Built-in phrase codes.  Must match with grammar file.
#define TYPE_EOF 0
#define TYPE_TAG 1
#define TYPE_STRING 2
#define TYPE_CHARCONST 3
#define TYPE_CHAR 4
#define TYPE_INT 5
#define TYPE_HEXINT 6
#define TYPE_KEYWORD 7

int bestparse = -1; // for error reporting.
char *looking_for = "<UNKNOWN>";    // 'while looking for <PHRASENAME>' (or literal) ...
/*>*/

/*< Data structures */

  /*< C[] is the source character token stream */

  typedef struct sourceinfo { // ATOMS for processed input stream
    char *s; // string contents
    int l;   // lineno
    int col; // column
    int t;   // type - tag, "string", 'charconst', or char, so far
    char *f; // source or includefile  name
  } sourceinfo;

  static sourceinfo *c = NULL;
  int nextfree = 0, arraysize = 0;
  char *onecharstr;
  /*>*/

  /*< A[] is the Analysis record.  Contents point to a C[] structure when the item is a BIP,
     otherwise the format is [phraseno] [altno] [no-of-subphrases] [subphrases and/or BIPs...]
     for example, if A[25] contained 10, that would mean that the token for A[25] was stored
     in C[10].  (And the C string would be at &stringpool[c[A[25]]]?) */
  static int *A = NULL;  /* Flex array, expanded using 'makespace' below */
  int next_free_a = 0, a_size = 0;
  /*>*/

  /* variables used by line-reconstruction (lexer) */
  FILE *sourcefile;
  char *curfile;
  int startline = TRUE, whitespace = TRUE, lineno = 1, col = 0, ch, peek;
  /*>*/

/*< Support procedures */

  /*< Incidental I/O support */
#define ishex(ch) (int)strchr("0123456789ABCDEFabcdef", ch)
int hextobin(char x)
{
  // Too crude, just a quick hack.  Seldom called.
  if (('0' <= x) && (x <= '9')) {        return x-'0';
  } else if (('A' <= x) && (x <= 'F')) { return x-'A'+10;
  } else if (('a' <= x) && (x <= 'f')) { return x-'a'+10;
  } else {                               return 0;
    // ASSERT: INTERNAL ERROR!
  }
}
void indent(int depth, FILE *f)
{
  int i;
  for (i = 0; i < depth; i++) fprintf(f, "  ");
}
/*>*/

  /*< Support proc for storage management */
static void *makespace_(void *c, int nextfree, int *arraysize, int objsize) {

  if ((c == NULL) || (*arraysize == 0) || (nextfree == 0)) { // TOO MANY TESTS!  Need to pick the right one!
    c = calloc(1024, objsize); *arraysize = 1024;
  } else if (nextfree >= *arraysize) {
    *arraysize = (*arraysize * 2) + 1;
    c = (void *)realloc(c, (*arraysize+1) * objsize); // 0:arraysize, inclusive. eg 0:15, 0:127 etc
    // this was causing a crash on the second call, on Windows/TCC.  c was NULL on first call.
  }
  if (c == NULL) {fprintf(stderr, "makespace: %s\n", strerror(errno)); exit(errno);}
  return c;
}
#define makespace(c, nextfree, arraysize) c = (typeof(c))makespace_(c, nextfree, &arraysize, sizeof(c[0]))
/*>*/

  /*< Support procs for storing lexical units */
void stores(char *s, int lineno, int col, int type, char *fname) {
  makespace(c, nextfree, arraysize);
  c[nextfree].s = s; c[nextfree].l = lineno; c[nextfree].col = col;
  c[nextfree].f = fname; c[nextfree].t = type;
  nextfree++;
}

void storec(int ch, int lineno, int col, int type, char *fname) {
  onecharstr[ch*2] = ch; onecharstr[ch*2+1] = '\0';  // convert char to 1-char string before saving.
  stores(&onecharstr[ch*2], lineno, col, type, fname);
}
/*>*/

  /*< simple proc to recognise if a token is a keyword */
int iskeyword(char *s) {
  int i;
  for (i = 0; i < MAX_KEYWORD; i++) if (strcmp(s, keyword[i]) == 0) return TRUE;
  return FALSE;
}
/*>*//*>*/

/*< Main procedures - the parser and the code generator (which embodies the grammar) */

/*< Line reconstruction, which for this language equates to lexing */
void line_reconstruction(void)
{
  /*< Pre-process input ready for parsing.  Tokens are stored in array C[] */
  for (;;) {
    ch = fgetc(sourcefile); if (ch == EOF) break;
    ch &= 255; // int, positive.

    peek = fgetc(sourcefile); ungetc(peek, sourcefile);

    if (isalpha(ch)) {
        /*< token or keyword */
        int nextfree = 0, strsize = 0, startcol = col;
        char *token = NULL;

        whitespace = FALSE;
        for (;;) {
          makespace(token, nextfree, strsize);
          if (isalpha(ch) || isdigit(ch) || (ch == '_')) {
            // digits and '_' allowed after 1st char.
            col++;
            token[nextfree++] = ch;
          } else {
            token[nextfree] = '\0'; ungetc(ch, sourcefile);
            break;
          }
          ch = fgetc(sourcefile);
        }
        stores(strdup(token), lineno, startcol, iskeyword(token) ? TYPE_KEYWORD : TYPE_TAG, curfile);
        free(token);
        /*>*/
    } else if (isdigit(ch)) {
        /*< Number */
        int nextfree = 0, numsize = 0;
        char *number = NULL;

        // Store as a string...
        whitespace = FALSE;
        for (;;) {
          makespace(number, nextfree, numsize);
          if (isdigit(ch)) {
            col++;
            number[nextfree++] = ch;
          } else {
            number[nextfree] = '\0'; ungetc(ch, sourcefile);
            break;
          }
          ch = fgetc(sourcefile);
        }
        stores(strdup(number), lineno, col, TYPE_INT, curfile);
        free(number);
	/*>*/
    } else switch (ch) {

    case '$':
      /*< Hex constant \$[0-9a-fA-F]+ */

      // Q: store the '$' in the string or not?
      whitespace = FALSE;
      col++;
      if (ishex(peek)) {
        int nextfree = 0, numsize = 0;
        char *number = NULL;

        for (;;) {
          makespace(number, nextfree, numsize);
          ch = fgetc(sourcefile);
          if (ishex(ch)) {
            col++;
            number[nextfree++] = ch;
          } else {
            number[nextfree] = '\0'; ungetc(ch, sourcefile);
            break;
          }
        }
        stores(strdup(number), lineno, col, TYPE_HEXINT, curfile);
        free(number);
      } else {
        // Warn: probably an error... should not be any naked '$' symbols.
        // If the error to be prined would have been a generic syntax
        // error at the same location, then maybe give a more informative
        // error message such as "Unexpected character '$' near: ..."
        // On the other hand the generic mechanism probably reports this
        // almsost as accurately.
        storec(ch, lineno, col++, TYPE_CHAR, curfile);
      }
      /*>*/
      break;

    case '\'': // Handle 'c' character constants
    case '"': // Handle "string"
      /*< literals */
      {
        int nextfree = 0, strsize = 0, quotech = ch;
        char *string = NULL;

        whitespace = FALSE;
        col++;
        for (;;) {
          ch = fgetc(sourcefile); // Newlines are allowed
          col++;
          makespace(string, nextfree, strsize);
          if (ch == '\\') {
            ch = fgetc(sourcefile); col++;
            if (ch == '\\') {           string[nextfree++] = ch;
            } else if (ch == '\'') {    string[nextfree++] = '\'';
            } else if (ch == '"') {     string[nextfree++] = '"';
            } else if (ch == 'n') {     string[nextfree++] = '\n';
            } else if (ch == 'r') {     string[nextfree++] = '\r';
            } else if (ch == 't') {     string[nextfree++] = '\t';
            } else if (ch == 'x') {
              int x, x1, x2;
              x1 = fgetc(sourcefile); col++;
              if (!ishex(x1)) {
                // WARN: Bad format
                continue;
              }
              x2 = fgetc(sourcefile); col++;
              if (!ishex(x2)) {
                // WARN: Bad format
                continue;
              }
              x = (hextobin(x1)<<4) | hextobin(x2);
              if (x == 0) {
                // WARN: embedded NUL in a string is asking for trouble...
              }
              string[nextfree++] = x;
            } else {
              // Warn of unknown (to me) \x escape.  Probably an error.
	      string[nextfree++] = '\\'; string[nextfree++] = ch;
            }
          } else if (ch != quotech) {   string[nextfree++] = ch;
          } else {
            string[nextfree] = '\0';
            break;
          }
        }

        if (quotech == '\'') {
          if (strlen(string) == 1) {
          } else if (strlen(string) <= 4) {
            // Warn that 'xx' as a 32-bit int is a non-standard extension
          } else {
            // Warn that this is probably a string with the wrong type of quote.
          }
        }
        stores(strdup(string), lineno, col, (quotech == '\'' ? TYPE_CHARCONST : TYPE_STRING), curfile);
        free(string);
      }
      break;
      /*>*/

    case '/':
      /*< COMMENTS (or just a divide symbol) */
      col++;
      whitespace = FALSE;
      if (peek == '/') {
        // Handle line comment
        do {ch = fgetc(sourcefile);} while (ch != '\n');
        lineno++; col = 0; whitespace = TRUE;
      } else if (peek == '*') {
        /* Handle potential multi-line comment */
        ch = fgetc(sourcefile); // Now we have read '/*'
        for (;;) {
          col++;
          ch = fgetc(sourcefile); peek = fgetc(sourcefile);
          if ((ch == '*') && (peek == '/')) break;
          ungetc(peek, sourcefile);
        }
        col += 2;
        (void)fgetc(sourcefile); // Remove '/'
        // QUESTION: How does this affect # directives? 
      } else {
        storec(ch, lineno, col, TYPE_CHAR, curfile);
      }
      break;
      /*>*/

    // WHITESPACE
    case '\n':      lineno++;
    case '\r':      startline = TRUE; col = 0; whitespace = TRUE;
      break;

    case '\t':
    case ' ':       col++; // Does not affect whitespace
      break;

    // DIRECTIVES
    case '#':
      // If we interpret any #-directives while lexing, we don't want to
      // do an expensive test on every token, so what we can do is set
      // a countdown timer on the introductory token (either this '#'
      // or the actual keyword such as 'ifdef') and then test that the
      // *previous* tokens match when the timer hits 0, eg
      // C[cp-3] == '#' && C[cp-2] == 'include' ... etc

      if (!whitespace) {
        // WARN: probably an error... should not be any '#' symbols in the
        // middle of a line. (This language uses "!=" or "<>" for not-equal)
      }
      // Drop through

    default:
      whitespace = FALSE;
      storec(ch, lineno, col++, TYPE_CHAR, curfile);
    }
  }
  // set up a dummy at the end because we sometimes look ahead by 1
  // in the parsing code and don't want to hit uninitialised data.
  c[nextfree].t = TYPE_EOF;
  c[nextfree].s = "<EOF>";
  c[nextfree].l = lineno;
  c[nextfree].col = col;
  /*>*/
}
/*>*/

/*< Abstract Syntax Tree data structures */
#define opsym(root) AST[root]
#define leftchild(root) AST[root+1]
#define rightchild(root) AST[root+2]
#define nthchild(root, n) AST[root+n]

typedef int TRIP; // A 'trip' was originally a 'triple' of <opcode, operand, operand>
                  // Now we use n-ary tuples but retain the name for nostalgia's sake...

#define MAXTRIPS 1024
int AST[MAXTRIPS];
int nexttrip = 0;

// We use a stringpool, and strings are indexes into this pool.  This
// is useful for the same reasons that the AST is an array indexed
// by integers rather than a struct with pointers.  It may also
// save space by reusing common strings.  And we get a free tag
// to describe strings.  (Also we can compare strings just with
// an integer tag comparison, if we ever want to)

// not everything uses this yet.  search for malloc and realloc to find potential problems

char stringpool[1024];
int nextstring = 0;

/*< Several tables describing the AST opcodes.  Some are only relevant to diagnostic output. */

// the first 3 tables are used by the compiler:

typedef enum opcode {
    /*< names of AST operators */
    ERROR, NOOP,
    SEQUENCE, DEFFN, DEFPARAM, PARAM,
    CONST, TAG, OPERATOR,
    ASSIGN, DECLARE, VAR, IFTHEN, IFTHENELSE,
    INPUT, PRINT, FNCALL,
    AND, OR,
    ADD, SUB, MUL, DIV,
    LSH, RSH, EXP,
    NEG, NOT, PAREN,
    EQ, NE, LT, GT, LE, GE,
    ASS,
    INT, REAL,
    LOOP, ENDLOOP, BREAK, CONTINUE,
    REDIRECT,
    LINE
    /*>*/
} OPCODE;

int prio[] = {
    /*< priority level of infix operators */
    0/*ERROR*/, 0/*NOOP*/,
    0/*SEQUENCE*/, 0/*DEFFN*/, 0 /*DEFPARAM*/, 0/*PARAM*/,
    0/*CONST*/, 0/*TAG*/, 0/*OPERATOR*/,
    0/*ASSIGN*/, 0/*DECLARE*/, 0/*VAR*/, 0/*IFTHEN*/, 0/*IFTHENELSE*/,
    0/*INPUT*/, 0/*PRINT*/, 0/*FNCALL*/,
    6/*AND*/, 5/*OR*/,
    3/*ADD*/, 3/*SUB*/, 4/*MUL*/, 4/*DIV*/,
    7/*LSH*/, 7/*RSH*/, 8/*EXP*/,
    9/*NEG*/, 9/*NOT*/, 0/*PAREN*/,
    2/*EQ*/, 2/*NE*/, 2/*LT*/, 2/*GT*/, 2/*LE*/, 2/*GE*/,
    1/*ASS*/,
    0/*INT*/, 0/*REAL*/,
    0/*LOOP*/, 0/*ENDLOOP*/, 0/*BREAK*/, 0/*CONTINUE*/,
    0/*REDIRECT*/,
    0/*LINE*/
    /*>*/
};

int arity[] = {
    /*< Number of operands in the tuple for this opcode.  Most are triples. */
    1, 1,
    3, 4 /* DEFFN */, 3 /* DEFPARAM*/, 3,
    3, 3, 3,
    3, 2 /* DECLARE */, 3, 3, 4,
    2, 2, 3 /* FNCALL */,                /* Variable! Store in the struct? */
    3, 3,
    3, 3, 3, 3,
    3, 3, 3,
    /* NEG */ 2, /* NOT */ 2, /* PAREN */2,
    3, 3, 3, 3, 3, 3,
    3,
    3, 3,
    3, 3, 3, 3,
    3,
    3
    /*>*/
};

// name was intended only for debugging, but in this initial implementation
// it is also being used as the source of mnemonics for the code generator.
// In real life, the machine opcodes would _not_ map 1:1 with AST operators.

char *name[] = {
    /*< ascii representation of AST operator names for debugging */
    "ERROR", "NOOP",
    "SEQUENCE", "DEFFN", "DEFPARAM", "PARAM",
    "CONST", "TAG", "OPERATOR",
    "ASSIGN", "DECLARE", "VAR", "IFTHEN", "IFTHENELSE",
    "INPUT", "PRINT", "FNCALL",
    "AND", "OR",
    "ADD", "SUB", "MUL", "DIV",
    "LSH", "RSH", "EXP",
    "NEG", "NOT", "PAREN",
    "EQ", "NE", "LT", "GT", "LE", "GE",
    "ASS",
    "INT", "REAL",
    "LOOP", "ENDLOOP", "BREAK", "CONTINUE",
    "REDIRECT",
    "LINE"
    /*>*/
};

// the following tables are only used for debugging:

char *shortname[] = {
    /*< ascii representation of AST operators for display when drawing trees */
    "ERROR", "%)",
    "';'", "DEFFN", "DEFPARAM", "','",
    "CONST", "TAG", "OPERATOR",
    "=", "DECLARE", "VAR", "'if (...) then'", "'if (...) then ... else ...'",
    "INPUT", "PRINT", "(%",
    "&&", "||",
    "'+'", "'-'", "'*'", "'/'",
    "<<", ">>", "'^'",
    "'-'", "'~'", "'()'",
    "'=='", "'!='", "'<'", "'>'", "'<='", "'>='",
    "'='",
    "INT", "REAL",
    "LOOP", "ENDLOOP", "BREAK", "CONTINUE",
    "REDIRECT",
    "LINE"
    /*>*/
};

char *c_infix_op[] = {
    /*< ascii representation of infix operators for display.  Non-infix ops have dummy values. */
    "ERROR", "NOOP",
    ";", "DEFFN", "DEFPARAM", ",",
    "CONST", "TAG", "OPERATOR",
    "=", "DECLARE", "VAR", "if (...) then", "if (...) then ... else ...",
    "INPUT", "PRINT", "FNCALL",
    "&&", "||",
    "+", "-", "*", "/",
    "<<", ">>", "^",
    "-", "~", "()",
    "==", "!=", "<", ">", "<=", ">=",
    "=",
    "INT", "REAL",
    "LOOP", "ENDLOOP", "BREAK", "CONTINUE",
    "REDIRECT",
    "LINE"
    /*>*/
};

int display_children[] = {
    /*< In tree-drawing code, how many children do we draw for this node?  Not always the same as the arity. */
    0, 0,
    2, 3 /* DEFFN */, 2 /* DEFPARAM */, 2,
    0, 0, 0,
    2, 1, 0, 2, 3,
    1, 1, 2 /* FNCALL */,                 /* Variable no of params */
    2, 2,
    2, 2, 2, 2,
    2, 2, 2,
    1, 1, 1,
    2, 2, 2, 2, 2, 2,
    2,
    1, 0,
    0/*LOOP*/, 0, 0, 0,
    0,
    0
    /*>*/
};

int display_leftchild[] = {
    /*< Do we display the left child of the node? */
    FALSE, FALSE,
    TRUE, FALSE, TRUE, TRUE,
    FALSE, FALSE, FALSE,
    TRUE, TRUE, FALSE, TRUE, TRUE,
    TRUE, TRUE, TRUE /* FNCALL */,
    TRUE, TRUE,
    TRUE, TRUE, TRUE, TRUE,
    TRUE, TRUE, TRUE,
    TRUE, TRUE, TRUE,
    TRUE, TRUE, TRUE, TRUE, TRUE, TRUE,
    TRUE,
    FALSE, FALSE,
    FALSE, FALSE, FALSE, FALSE,
    TRUE,
    FALSE
    /*>*/
};

int display_rightchild[] = {
    /*< Do we display the right child of the node? */
    FALSE, FALSE,
    TRUE, FALSE, TRUE, TRUE,
    FALSE, FALSE, FALSE,
    TRUE, FALSE, FALSE, TRUE, TRUE,
    FALSE, FALSE, TRUE /* FNCALL */,     /* Variable no of args >= 0 */
    TRUE, TRUE,
    TRUE, TRUE, TRUE, TRUE,
    TRUE, TRUE, TRUE,
    FALSE, FALSE, FALSE,
    TRUE, TRUE, TRUE, TRUE, TRUE, TRUE,
    TRUE,
    FALSE, FALSE,
    FALSE, FALSE, FALSE, FALSE,
    FALSE,
    FALSE
    /*>*/
};
/*>*/


char *nameof(TRIP t)
{
    /*< Extract the name of a variable or the value of a const */
    // I realise that the malloc here is heap lossage, but for now
    // its not worth fixing.  Choices are either use a stack to allocate
    // temporary space (same as in tacc) or keep a cyclic buffer, since
    // this is only called in the context of printing so we really
    // only need a few active at once.
    char *result = malloc(40 /* constants are also bad. I'm lazy. */);
    int op = opsym(t);
    if (op == CONST) {
	sprintf(result, "%d", rightchild(t));
    } else if (op == VAR) { // for simpler diagrams, skip a level
	sprintf(result, "%s",
		stringpool+rightchild(leftchild(t))); // punt to tag
    } else if (op == TAG) { // for simpler diagrams, skip a level
	sprintf(result, "%s",
		stringpool+leftchild(t));
    } else {
	sprintf(result, "%s", shortname[op]);
    }
    return(result);
    /*>*/
}

#define tripsize(i) arity[opsym(i)]

void printtrip(TRIP i);

TRIP mkop(OPCODE op) {
    return make_binary_tuple(OPERATOR, op, -1); // rightchild could be ptr to string?
}

// note we're taking a slight shortcut here by storing the opcode directly in the
// first field.  If fact a cleaner design might have been to make that field point to
// an 'opcode' trip, which is a unary operator whose sole parameter is the opcode.
// this would require one extra indirection at the point of access, which could
// be hidden in a procedure call anyway.

TRIP make_unary_tuple(OPCODE op, TRIP parm1) {
    /*< Create a tuple for a unary operator */
    int trip = nexttrip;
    opsym(trip) = op;  nexttrip += tripsize(trip);
    leftchild(trip) = parm1;
#ifdef DEBUG_TRIPS_DURING
    printtrip(trip);
#endif
    return trip;
    /*>*/
}

TRIP make_binary_tuple(OPCODE op, TRIP parm1, TRIP parm2) {
    /*< Create a tuple for a binary operator */
    int trip = nexttrip;
    opsym(trip) = op;  nexttrip += tripsize(trip);
    leftchild(trip) = parm1;
    rightchild(trip) = parm2;
#ifdef DEBUG_TRIPS_DURING
    printtrip(trip);
#endif
    return trip;
    /*>*/
}

TRIP make_nary_tuple(OPCODE op, TRIP parm1, ...) {
    /*< Create a tuple for an n-ary operator.  Uses stdargs for arbitrary no. of params */
    // we use this for if/then/else, and procedure calls with parameter lists.
    int trip;
    int parm;
    va_list ptr;

    trip = nexttrip++;
    opsym(trip) = op;
    leftchild(trip) = parm1;
    va_start(ptr, parm1);
    for (parm = 2; parm <= arity[op]; parm++) {
	nthchild(trip, parm) = va_arg(ptr, TRIP); nexttrip += 1;
    }
    va_end(ptr);
#ifdef DEBUG_TRIPS_DURING
    printtrip(trip);
#endif
    return trip;
    /*>*/
}

TRIP make_proc_name(int pooloffset) {
    return make_binary_tuple(TAG, 0, pooloffset);
}

TRIP make_int_const(int datatype, char *value) {
    return make_binary_tuple(CONST, datatype, (int)atol(value));
}

TRIP make_real_const(int datatype, char *value) {
    return make_binary_tuple(CONST, datatype, (int)atof(value) /* Hacky.  Should fix this. */);
}

TRIP getvar(char *s) { /* tag must exist */
    /*< Look up a tag in the string pool. */
    /* getvar needs to be more complex.  Currently it just maps
       from a string to the triple for a pre-declared var, adding
       any unrecognised string as an int.  Should really add it as
       an error type, and handle scope rules */

    int i, trip, tag;

    strcpy(stringpool+nextstring, s); /* Create a backstop in case not found */
    for (tag = 0; tag <= nextstring; tag++) {
	if (strcmp(stringpool+tag, s) == 0) {
	    break; /* found, one way or another */
	}
    }
    if (tag == nextstring) {                               /* Not found, auto-declare it */
	nextstring += strlen(s)+1;
	trip = make_binary_tuple(TAG, 0, tag);
	return make_binary_tuple(VAR, trip, INT);
    }

    /* having located the stringpool entry,
       now find the appropriate declaration whose tag is using it */
    for (i = 0; i < nexttrip; i++) {
        trip = leftchild(i);
	if ((opsym(i) == VAR) && (rightchild(i) == INT) &&
	    (opsym(trip) == TAG) && (rightchild(trip) == tag)
           ) return i;
    }
    fprintf(stderr, "Oops: cannot find declaration for this name!\n"); exit(1);
    /*>*/
}

int str_to_pool(char *s)
{
    int tag;
    for (tag = 0; tag <= nextstring; tag++) {
	if (strcmp(stringpool+tag, s) == 0) {
	    return tag; /* found, one way or another */
	}
    }
}

TRIP newtag(char *s) { /* tag must *not* exist */
    /*< Create a new tag and add it to the stringpool. */
    int i, trip, tag;

    strcpy(stringpool+nextstring, s); /* Create a backstop for when not found */
    tag = str_to_pool(s);
    if (tag != nextstring) {
	fprintf(stderr, "ERROR: name '%s' already exists.\n", s);
	/* forget about scope rules for now */
    } else nextstring += strlen(s)+1; /* Not found, add it */
    /*>*/
    return make_binary_tuple(TAG, 0, tag);
}
/*>*/

/*< Debugging */
#if defined(DEBUG_TRIPS_AFTER) || defined(DEBUG_TRIPS_DURING)
/*< Diagnostic procedure to print a triple. */
/*< This section really isn't very interesting.  It's not structured for
   folding, and it's not relevant to the compiler algorithms.
   You can comfortably skip this part of the code... */
/* this is ONLY used for diagnostics and is the same in all test harness programs */
void printtrip(TRIP i) {
  int op, parm1, parm2;

  if (i < 0) {
    fprintf(stdout, "** Too small: TRIP %d\n", i); (void)*(int *)0;
  }

  if (i >= MAXTRIPS) {
    fprintf(stdout, "** Too large: TRIP %d\n", i); (void)*(int *)0;
  }

  op = opsym(i);
  parm1 = leftchild(i);
  parm2 = rightchild(i);

  switch (opsym(i)) {

  case DECLARE:
    fprintf(stdout, "// %d: %s [@AST %d]\n",
      i, name[op], parm1);
    break;

  case VAR:
    fprintf(stdout, "// %d: %s %s[@AST %d] TYPE=%s\n",
      i, name[op], stringpool+rightchild(parm1), parm1, name[parm2]);
    break;

  case TAG:  /* parm1 not used for the moment */
    fprintf(stdout, "// %d: %s %s[@Stringpool %d]\n",
      i, name[op], stringpool+parm2, parm2);
    break;

  case DEFFN:
    fprintf(stdout, "// %d: %s \"%s\" [arglist @AST %d] [fnbody @AST %d]\n",
      i, name[op], stringpool+rightchild(leftchild(parm1)), parm2, nthchild(i, 3));
    break;

  case CONST:
    fprintf(stdout, "// %d: %s %s %d\n",
      i, name[op], name[parm1], parm2);
    break;

  case OPERATOR:
    fprintf(stdout, "// %d: %s %s  (Orphaned AST cell)\n",
      i, name[op], name[parm1]);
    break;

  case ASSIGN:
    fprintf(stdout, "// %d: %s [declared @AST %d] [value @AST %d]\n",
      i, name[op], parm1, parm2);
    break;

  case IFTHENELSE:
    fprintf(stdout, "// %d: %s [condition @AST %d] [then-statement @AST %d] [else-statements @AST %d]\n",
      i, name[op], parm1, parm2, nthchild(i, 3));
    break;

  case IFTHEN:
    fprintf(stdout, "// %d: %s [condition @AST %d] [then-statement @AST %d]\n",
      i, name[op], parm1, parm2);
    break;

  case LINE:
    fprintf(stdout, "// %d: %s %d: %s\n",
      i, name[op], parm1, (char *)parm2);
    break;

  case AND:
  case OR:
  case ADD:
  case SUB:
  case MUL:
  case DIV:
  case LSH:
  case RSH:
  case EXP:
    fprintf(stdout, "// %d: %s [@AST %d] [@AST %d]\n",
      i, name[op], parm1, parm2);
    break;

  case EQ:
  case NE:
  case LT:
  case GT:
  case LE:
  case GE:
    fprintf(stdout, "// %d: [@AST %d] %s [@AST %d]\n",
      i, parm1, name[op], parm2);
    break;


  case NEG:
  case NOT:
    fprintf(stdout, "// %d: %s [@AST %d]\n",
      i, name[op], parm1);
    break;

  default:
    if (arity[opsym(i)] == 3) {
      fprintf(stdout, "// %d: %s %d %d\n",
        i, name[op], parm1, parm2);
    } else if (arity[opsym(i)] == 2) {
      fprintf(stdout, "// %d: %s [@AST %d]\n",
        i, name[op], parm1);
    } else {
      fprintf(stdout, "// %d: %s ...?  (%d ops)\n",
        i, name[op], arity[opsym(i)]);
    }
    break;
  }
}
/*>*//*>*/
#endif

#ifdef DEBUG_TREES
/*< Ascii-art debugging procedure for drawing trees */
/*< OK, look, you really don't want to expand this section.  It's not structured for
   folding, and it's not relevant to the compiler algorithms.  Trust me, you can skip
   this... (and did I mention it was huge and monolithic???) */

int del = 4/*was 1*/; /* distance of graph columns */
int eps = 3; /* distance of graph lines */

/* interface for drawing (can be replaced by "real" graphic using GD or other) */
void graphInit (void);
void graphFinish();
void graphBox (char *s, int *w, int *h);
void graphDrawBox (char *s, int c, int l);
void graphDrawArrow (int c1, int l1, int c2, int l2);

/* recursive drawing of the syntax tree */
void exNode (int trip, int c, int l, int *ce, int *cm, int depth, int *needed);

/*****************************************************************************/

/* main entry point of the manipulation of the syntax tree */

/* draw_tree is taken from the yacc/lex demo by Thomas Niemann
   at http://epaperpress.com/lexandyacc/

   - I prefer my own drawing code whenever possible (the trees look
     prettier) but my code does not handle n-tuples where n > 2,
     so I fall back to this code if a diagram requires such a tuple. */

void draw_tree_orig(int root);

void dottree(int root)
{
  char *operator;
  TRIP leftkid, rightkid;
  int linkno = 0;

  operator = nameof(root);
  leftkid = leftchild(root);
  rightkid = rightchild(root);

  fprintf(stdout, "\"node%d\" [\n   label = \"<f0> %d: %s ", root, root, operator);
  if (display_leftchild[opsym(root)]) fprintf(stdout, "| <f%0d> %d: ", ++linkno, root+1);
  if (display_rightchild[opsym(root)]) fprintf(stdout, "| <f%0d> %d: ", ++linkno, root+2);
  linkno = 0;

  fprintf(stdout, "\"\n   shape = \"record\"\n];\n\n");

  if (display_leftchild[opsym(root)]) dottree(leftkid);
  if (display_rightchild[opsym(root)]) dottree(rightkid);

  if (display_leftchild[opsym(root)])
      fprintf(stdout, "\"node%0d\":f1 -> \"node%0d\":f0 [\n id = %d\n];\n\n",
              root, leftchild(root), 888);
  if (display_rightchild[opsym(root)])
      fprintf(stdout, "\"node%0d\":f2 -> \"node%0d\":f0 [\n id = %d\n];\n\n",
	      root, rightchild(root), 999);
}

void draw_tree(int trip) {
    int rte, rtm, needed;

    fprintf(stdout, "\nTree for AST[%d]:\n", trip);

    graphInit ();
    needed = FALSE;
    exNode (trip, 0, 0, &rte, &rtm, 0, &needed);
    if (needed) {
      graphFinish();
    } else {
      draw_tree_orig(trip);
    }

#ifdef DOT_SUPPORT
    fprintf(stdout, "digraph g {\n");
    fprintf(stdout, "graph [\n");
    fprintf(stdout, "   rankdir = \"LR\"\n");
    fprintf(stdout, "];\n\n");
    fprintf(stdout, "node [\n");
    fprintf(stdout, "   fontsize = \"16\"\n");
    fprintf(stdout, "   shape = \"ellipse\"\n");
    fprintf(stdout, "];\n\n");
    fprintf(stdout, "edge [\n");
    fprintf(stdout, "];\n\n");
    dottree(trip);
    fprintf(stdout, "}\n");
#endif

}

/*c----cm---ce---->                       drawing of leaf-nodes
 l leaf-info
 */

/*c---------------cm--------------ce----> drawing of non-leaf-nodes
 l            node-info
 *                |
 *    -------------     ...----
 *    |       |               |
 *    v       v               v
 * child1  child2  ...     child-n
 *        che     che             che
 *cs      cs      cs              cs
 *
 */

void indentsp(int d)
{
  int i;
  for (i = 0; i < d*4; i++) {
    putchar(' ');
  }
}

void exNode
    (   int trip,
        int c, int l,        /* start column and line of node */
        int *ce, int *cm,    /* resulting end column and mid of node */
        int depth, int *needed
    )
{
    int op;
    int w, h;           /* node width and height */
    char *s;            /* node text */
    int cbar;           /* "real" start column of node (centred above subnodes) */
    int k;              /* child number */
    int che, chm;       /* end column and mid of children */
    int cs;             /* start column of children */
    char word[40];

//indentsp(depth);fprintf(stdout, "start: TRIP=%d  startcol=%d  startline=%d\n", trip, c, l);

    if (trip == -1) return;
    op = opsym(trip);
    if (display_children[op] >= 3) *needed = TRUE;

    s = nameof(trip);
    sprintf(word, "%s", s);
    s = word;

//indentsp(depth);fprintf(stdout, "graphbox: s = %s\n", s);

    /* construct node text box */
    graphBox (s, &w, &h);
    cbar = c;
//indentsp(depth);fprintf(stdout, "assign: c=%d\n", c);
    *ce = c + w;
    *cm = c + w / 2;

    /* node is leaf */
    if (
         (op == CONST ||
          op == VAR ||
          display_children[op] == 0)
       ) {
//indentsp(depth);fprintf(stdout, "drawbox: s = %s  cbar=%d\n", s, cbar);
        graphDrawBox (s, cbar, l);
        return;
    }

    /* node has children */
    cs = c;
//indentsp(depth);fprintf(stdout, "node has %d children: cs=c=%d\n", display_children[op], c);
    for (k = 1; k <= display_children[op]; k++) {
//indentsp(depth);fprintf(stdout, "%d: exnode1 %d  cs=%d\n", k, nthchild(trip, k), cs);
        exNode (nthchild(trip, k), cs, l+h+eps, &che, &chm, depth+1, needed);
        cs = che;
    }

    /* total node width */
    if (w < che - c) {
        cbar += (che - c - w) / 2;
        *ce = che;
        *cm = (c + che) / 2;
    }

    /* draw node */
//indentsp(depth);fprintf(stdout, "cbar=%d\n", cbar);
    graphDrawBox (s, cbar, l);

    /* draw arrows (not optimal: children are drawn a second time) */
    cs = c;
    for (k = 1; k <= display_children[op]; k++) {
//indentsp(depth);fprintf(stdout, "%d: exnode2 %d  cs=%d\n", k, nthchild(trip, k), cs);
        exNode (nthchild(trip, k), cs, l+h+eps, &che, &chm, depth+1, needed);
        graphDrawArrow (*cm, l+h, chm, l+h+eps-1);
        cs = che;
    }
}

/* interface for drawing */

#define lmax 200
#define cmax 200

char graph[lmax][cmax]; /* array for ASCII-Graphic */

void graphTest (int l, int c)
{   int ok;
    ok = 1;
    if (l < 0) ok = 0;
    if (l >= lmax) ok = 0;
    if (c < 0) ok = 0;
    if (c >= cmax) ok = 0;
    if (ok) return;
    printf ("\n+++error: l=%d, c=%d not in drawing rectangle 0, 0 ... %d, %d", 
        l, c, lmax, cmax);
//    fprintf (stderr, "\n+++error: l=%d, c=%d not in drawing rectangle 0, 0 ... %d, %d", 
//        l, c, lmax, cmax);
    {
    int i, j;
    int lmx=20, cmx=60;
    for (i = 0; i < lmx; i++) {
        for (j = cmx-1; j > 0 && graph[i][j] == ' '; j--);
        graph[i][cmx-1] = 0;
        if (j < cmx-1) graph[i][j+1] = 0;
        if (graph[i][j] == ' ') graph[i][j] = 0;
    }
    for (i = lmx-1; i > 0 && graph[i][0] == 0; i--);
    printf ("\n");
    for (j = 0; j <= i; j++) printf ("\n    // %s", graph[j]);
    printf("\n");
    };
    exit (1);
}

void graphInit (void) {
    int i, j;
    for (i = 0; i < lmax; i++) {
        for (j = 0; j < cmax; j++) {
            graph[i][j] = ' ';
        }
    }
}

void graphFinish() {
    int i, j;
    char *s;
    for (i = 0; i < lmax; i++) {
        for (j = cmax-1; j > 0 && graph[i][j] == ' '; j--);
        graph[i][cmax-1] = 0;
        if (j < cmax-1) graph[i][j+1] = 0;
        if (graph[i][j] == ' ') graph[i][j] = 0;
    }
    for (i = lmax-1; i > 0 && graph[i][0] == 0; i--);
    printf ("\n");

    for (j = 0; j <= i; j++) {
      char *p;
      s = graph[j];
      // hacks to slightly improve formatting
      if (j == 0) s += 2;
      else if (j == 1) s += 1;
      else if ((p=strchr(s, '|')) == NULL || p[1]=='|') s += 2;
      printf ("\n    // %s", s);
    }
    printf("\n\n");
}

void graphBox (char *s, int *w, int *h) {
    *w = strlen (s) + del;
    *h = 1;
}

void graphDrawBox (char *s, int c, int l) {
    int i;
//fprintf(stdout, "c=%d strlen=%d del=%d\n", c, strlen(s), del);
    graphTest (l, c+strlen(s)-1+del);
    for (i = 0; i < strlen (s); i++) {
        graph[l][c+i+del] = s[i];
    }
}

void graphDrawArrow (int c1, int l1, int c2, int l2) {
    int m;
    graphTest (l1, c1);
    graphTest (l2, c2);
    m = (l1 + l2) / 2;
    while (l1 != m) { graph[l1][c1] = '|'; if (l1 < l2) l1++; else l1--; }
    while (c1 != c2) { graph[l1][c1] = '-'; if (c1 < c2) c1++; else c1--; }
    while (l1 != l2) { graph[l1][c1] = '|'; if (l1 < l2) l1++; else l1--; }
    graph[l1][c1] = '|';
}

#endif
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
#ifdef DEBUG_TREES
/* See drawtree.c for test harness.  Small mods made for this interface. */
/* NOTE!!! Now we have n-ary trees, this will not work.  ifthenelse breaks */
static int tree_debug = (0!=0);
static int vertical = (0==0);
static int horizontal = (0==0);
static int wide = (0!=0);
static int trim = (0==0);
static int testone = (0==0);

//       row  col
long pic[256][256]; // 0..255 is char, >= 256 is ptr to string


int oldtextblit(int row, int col, char *src)
{
  // post-processing string expansion
  int l = 0;
  for (;;) {
    if (*src == '\0') break;
    if (tree_debug) fprintf(stderr, "1: Planting '%c' at [%d][%d]\n", *src, row, col);
    pic[row][col++] = *src++;
    l += 1;
  }
  return l;
}

int textblit(int row, int col, char *src)
{
  // store pointer to string, unpack later on output
  int l = strlen(src);
  pic[row][col] = (int)src;
  return (l+(wide?3:1))>>(wide?2:1); // half size because on diagonal
}


void layout(int id, int idx, int rowoff, int coloff, int *depth, int *width)
{
  char *operator;
  int op;
  int leftkid, rightkid;
  int leftkiddepth = 0, leftkidwidth = 0;
  int rightkiddepth = 0, rightkidwidth = 0;
  int deltadepth = 0, deltawidth = 0;
  int i;

  if (tree_debug) fprintf(stderr, ">> %d:layout(%d, rowcol[%d][%d], depth %d, width %d);\n", id, idx, rowoff, coloff, *depth, *width);

  operator = nameof(idx);
  leftkid = leftchild(idx);
  rightkid = rightchild(idx);

  // Anchor the corner node.
  (void)textblit(rowoff, coloff, operator); /* not strcpy - don't copy NUL */
  deltawidth = 1;
  if (display_rightchild[opsym(idx)]) {
    int len = ((strlen(nameof(leftkid))+(wide?3:1))>>(wide?2:1))+1; // text on the diagonal
    while (len-- > 1) {deltawidth += 1; pic[rowoff][coloff-1+deltawidth] = (vertical ? '\\' : '-');}
    // attach the RHS tree
    if (tree_debug) fprintf(stderr, "Recursing to right node %d\n", rightkid);
    layout(2*id, rightkid, rowoff, coloff+deltawidth, &rightkiddepth, &rightkidwidth);
    deltadepth = rightkiddepth;
  } else {
    deltadepth = 1; /* The op itself */
  }
// testing: correcting a typo
  if (((strlen(operator)+(wide?3:1))>>(wide?2:1)) >= deltawidth) deltawidth = ((strlen(operator)+(wide?3:1))>>(wide?2:1))+2;

  if (display_leftchild[opsym(idx)]) {
    // draw the down link

    // calculate extra height
    if ((((strlen(nameof(leftkid))+(wide?3:1))>>(wide?2:1))) > deltadepth) {
      deltadepth = ((strlen(nameof(leftkid))+(wide?3:1))>>(wide?2:1));
    }

    for (i = 1; i < deltadepth+1 /* +1 for spacer row */; i++) {
      if (tree_debug) fprintf(stderr, "2: Planting '%c' at [%d][%d]\n", '/', rowoff+i, coloff);
      pic[rowoff+i][coloff] = (horizontal ? '/' : '|');
    }
    // recurse on the LHS tree
    if (tree_debug) fprintf(stderr, "Recursing to left node %d\n", leftkid);
    layout(2*id+1, leftkid, rowoff+deltadepth+1, coloff, &leftkiddepth, &leftkidwidth);
    *depth = (*depth) + leftkiddepth + deltadepth + 1;
  } else *depth = (*depth) + deltadepth;

  if (rightkidwidth+deltawidth > leftkidwidth) {
    *width = (rightkidwidth+deltawidth);
  } else {
    *width = leftkidwidth;
  }

  if (tree_debug) fprintf(stderr, "<< %d:layout(%d, rowcol[%d][%d], depth %d, width %d);\n", id, idx, rowoff, coloff, *depth, *width);
}

void draw_tree_orig(int root)
{
  int depth = 0, width = 0, row, col, offset, trimmable;

  fprintf(stdout, "\n");
  // Init.
  for (col = 0; col < 256; col++) {
    for (row = 0; row < 256; row++) {
      pic[row][col] = ' ';
    }
  }

  /* Generate layout */
  layout(1, root, 128, 0, &depth, &width);

  if (tree_debug) fprintf(stderr, "Dump layout: rows = %d cols = %d\n", depth, width);
  if (tree_debug) fflush(stderr);

  if (vertical) {
    /* apply vertical shear first */
    offset = 1;
    for (col = 1; col < 256; col++) {
      // move this column down by 'offset'
      for (row = 255; row > offset; row--) {
        pic[row][col] = pic[row-offset][col]; pic[row-offset][col] = ' ';
      }
      offset += 1;
    }
  }

  if (horizontal) {
    /* apply horizontal shear next */
    row = 255;  // start at bottom of drawing
    offset = 0;
    for (;;) {
      static long temp[1024];
      for (col = 0; col < 256; col++) {
        temp[col] = ' ';
      }
      for (col = 0; col < 256; col++) {
        temp[col*2+offset] = pic[row][col];
        temp[col*2+offset+1] = ' ';
      }
      for (col = 0; col < 256; col++) {
        pic[row][col] = temp[col];
      }
      if (row == 0) break;
      offset += 1; /* more shear on next row up */
      row -= 1;
    }
  }

  if (trim) {
    trimmable = (0==0);
    for (;;) {
      for (row = 0; row < 256; row++) {
        if (pic[row][0] != ' ') {
          trimmable = (0!=0);
          break;
        }
      }
      if (!trimmable) break;
      for (row = 0; row < 256; row++) {
        for (col = 0; col+1 < 256; col++) {
          pic[row][col] = pic[row][col+1];
        }
        pic[row][255] = ' ';
      }
    }
  }

  if (wide) {
    /* apply widening last */
    row = 255;  // start at bottom of drawing
    offset = 0;
    for (;;) {
      static long temp[1024];
      for (col = 0; col < 256; col++) {
        temp[col] = ' ';
      }
      for (col = 0; col < 256; col++) {
        temp[col*2+offset] = pic[row][col];
        temp[col*2+offset+1] = ' ';
      }
      for (col = 0; col < 256; col++) {
        pic[row][col] = temp[col];
      }
      if (row == 0) break;
      row -= 1;
    }
  }

  /* display tree */
  for (row = 0; row < 256; row++) {
    trimmable = (0 == 0);
    for (col = 0; col < 256; col++) {
      if (pic[row][col] != ' ') {
        trimmable = (0!=0);
        break;
      }
    }
    if (!trimmable) {
      fprintf(stdout, "  ");  // INDENT
      for (col = 255; col >= 0; col--) {
        if (pic[row][col] != ' ') break;
        pic[row][col] = '\0';
      }
      printf("  //    ");
      for (col = 0; col < 256; col++) {
        if ((pic[row][col] < -128) || (pic[row][col] > 255)) {
          oldtextblit(row, col, (char *)pic[row][col]);
        } else if (pic[row][col] == '\0') break;
        putchar(pic[row][col]);
      }
      putchar('\n');
    }
  }
  putchar('\n');
  fflush(stdout);
  return;
}
/*>*//*>*/
#endif /* DEBUG_TREES */
/*>*/

/*< Parsing.  (Table driven top-down recursive-descent parser) */

/*<  The parser used here is based on a design by Tony Brooker which was
    originally used in Atlas Autocode and the "Compiler Compiler".
    It generates a concrete syntax tree rather than the abstract
    syntax tree more popular in modern compilers.  A later phase
    converts from concrete to abstract.

    Note that the parsing procedure here is just a piece of code
    to walk a pre-built table.  There is nothing in this section
    which reflects the grammar, if that is what you are looking for.
    You'll find the grammar embedded in the 'compile()' procedure
    in the following section.
 */

int cp = 0;  // code pointer.  Has to be global state.
int ap = 0;  // Analysis record pointer.

int parse(int pp, int depth) // depth is only for indentation in diags
{
  /*< Main parsing procedure.  This is a table-driven parser, with the tables
      being generated from the grammar rules embedded in the 'compile' procedure
      below.  The result of the parse is a tree structure, and the values of the
      nodes in the tree structure are used to drive a large 'case' statement
      which selects the actions to be performed after a successful parse.

      There is no grammatical structure embedded in this procedure.  If you're
      looking for the grammar definition, see the procedure called 'compile' instead.
   */
  int saved_cp, saved_ap, i, gp, alts, match;

  /*< Initialisation */
  gp = phrase_start[pp-512-MAX_BIP];
  alts = gram[gp];

  /*< Debugging */
#ifdef DEBUG_PARSER
  if (debug_parser) {
    fprintf(stdout, "\n");
    indent(depth, stdout);
    fprintf(stdout, "Phrase %s/%d (%d alternatives) = ", phrasename[pp-512], pp, alts);
    fflush(stdout);
  }
#endif
  /*>*/
  gp++; // gp now points to first element (length) of first alt

  saved_cp = cp;
  saved_ap = ap;
  /*>*/

  for (i = 0; i < alts; i++) {
    /*< Starting with the root phrase, recursively examine each alternative */
    int each, phrases = gram[gp++], phrase_count, gap = 0;

    cp = saved_cp;
    ap = saved_ap;

    if (ap+3 > next_free_a) next_free_a = ap+3;
    makespace(A, next_free_a, a_size);

    A[ap++] = pp;   // record which phrase (could be done outside loop)
    A[ap++] = i;    // and which alt.


    // Count slots needed.  *Could* be precalculated and stored
    // in the grammar, either embedded (after the ALT) or as a
    // separate table

    for (each = 0; each < phrases; each++) if (gram[gp+each] >= 512) gap++;

    A[ap++] = gap;    // Count of alts (gap)
    // ap+gap now points to the slot after the space required, which
    // is where the first subphrase will be stored.
    ap = ap+gap; // recursive subphrases are stored after this phrase.
                 // ap is left updated if successful.


    // successfully parsed phrases are stored in A[saved_ap+3+n]

    if (saved_ap+3+gap > next_free_a) next_free_a = saved_ap+3+gap;
    makespace(A, next_free_a, a_size);

    /*< Debug */
    // this loop is only for diagnostics
#ifdef DEBUG_PARSER
    if (debug_parser) {
      fprintf(stdout, "\n");
      indent(depth, stdout);
      fprintf(stdout, "Alternative %d: (%d phrases) ", i+1, phrases);
      for (each = 0; each < phrases; each++) {
        int phrase = gram[gp+each];
        if (phrase < 256) {
          fprintf(stdout, " '%c'", phrase);
        } else if (phrase < 512) {
          fprintf(stdout, " \"%s\"/%d", keyword[phrase-256], phrase-256);
        } else if (phrase < 512+MAX_BIP) {
          fprintf(stdout, " {%s/BIP%d}", phrasename[phrase-512], BIP[phrase-512]);
        } else {
          fprintf(stdout, " <%s/%d>", phrasename[phrase-512], phrase);
        }
      }
      fprintf(stdout, "\n");
      fflush(stdout);
    }
#endif
    /*>*/

    match = TRUE; // stays true if all subphrases match
    phrase_count = 0; // only phrases which make it into the A record,
                      // i.e. excluding literals and keywords
    for (each = 0; each < phrases; each++) {
      /*< Within a single grammar rule (alternative), ensure that each subphrase is present */
      int phrase = gram[gp+each];

      /*< Debug */
#ifdef DEBUG_PARSER
      if (debug_parser) {
        indent(depth, stdout);
        fprintf(stdout, "Input token stream = '%s' '%s' '%s' ...\n",
          (cp < nextfree ? c[cp].s : "EOF"),
          (cp+1 < nextfree ? c[cp+1].s : "EOF"),
          (cp+2 < nextfree ? c[cp+2].s : "EOF"));
      }
#endif
      if (cp > bestparse) {
        static char s[128];
#ifdef DEBUG_PARSER
        if (phrase < 256) {
          sprintf(s, "'%c'", phrase);
        } else if (phrase < 512) {
          sprintf(s, "\"%s\"", keyword[phrase-256]);
        } else if (phrase < 512+MAX_BIP) {
          sprintf(s, "{%s}", phrasename[phrase-512]);
        } else {
          sprintf(s, "<%s>", phrasename[phrase-512]);
        }
#endif
        looking_for = s;
        bestparse = cp;
      }
#ifdef DEBUG_PARSER
      if (debug_parser) indent(depth, stdout);
#endif
      /*>*/

      if (phrase < 256) {
	/*< Literal */
#ifdef DEBUG_PARSER
        if (debug_parser) fprintf(stdout, "'%c'", phrase);
#endif
        if ((c[cp].t != TYPE_CHAR) || (c[cp].s[0] != phrase)) match = FALSE; else cp++;
        // Don't record literals
        /*>*/
      } else if (phrase < 512) {
	/*< Keyword */
#ifdef DEBUG_PARSER
        if (debug_parser) fprintf(stdout, "\"%s\"/%d", keyword[phrase-256], phrase-256);
#endif
        if (strcmp(keyword[phrase-256], c[cp].s) != 0) match = FALSE; else cp++;
        // Don't record keywords
        /*>*/
      } else if (phrase < 512+MAX_BIP) {
	/*< Built-in phrase */
        int where = ap; // next phrase to be parsed will be stored at current 'ap'.
#ifdef DEBUG_PARSER
        if (debug_parser) fprintf(stdout, "{%s/BIP%d}", phrasename[phrase-512], BIP[phrase-512]);
#endif
        if (c[cp].t != BIP[phrase-512]) match = FALSE; else {
          A[ap++] = phrase;
          A[ap++] = 1;
          A[ap++] = 1;
          A[ap++] = cp++;
          A[saved_ap+3+phrase_count++] = where; // Record BIP
        }/*>*/
      } else {
	/*< Recursive call to parser for a subphrase */
        int where = ap; // next phrase to be parsed will be stored at current 'ap'.
#ifdef DEBUG_PARSER
        if (debug_parser) fprintf(stdout, "<%s/%d>", phrasename[phrase-512], phrase);
#endif
        if (!parse(phrase, depth+1)) match = FALSE; else {
          A[saved_ap+3+phrase_count++] = where;
        }
        /*>*/
      }
      /*< debug */
#ifdef DEBUG_PARSER
      if (debug_parser) {
        fprintf(stdout, "\n");
        indent(depth, stdout);
        fprintf(stdout, "Tried alternative %d: result was %s\n", each+1, (match ? "TRUE" : "FALSE"));
        fflush(stdout);
      }
#endif
      /*>*/
      if (!match) break;
      /*>*/
    }
    gp += phrases; // move over all phrases, to next alt

    if (match) break;
#ifdef DEBUG_PARSER
    else if (debug_parser) {
      indent(depth, stdout);
      fprintf(stdout, "** Alternative %d FAILED.\n", i+1);
    }
#endif
    // gp now points to first element (length) of next alt, or start of next phrase
    /*>*/
  }

  return(match);
  /*>*/
}
/*>*//*>*/

/*< Convert the concrete syntax tree into an abstract syntax tree. */

/*< This is primarily the main 'compile()' procedure, which is actually where
   the grammar of the language is defined.  The grammar is extracted from
   in-line comments, and converted into a table by the 'takeon' program
   which you can find in the same directory as this file.

   (You can view the extracted grammar in file "varcalc.g")

   The style of compiler on which this design is based actually goes
   directly from concrete syntax tree to code generation - but that was
   from the days when memory was tight.  Since most modern compilers -
   and especially books about them - are AST-based, we'll take that
   extra step here in order to give our students an AST-based compiler
   to experiment with.
 */

/*< Compiler tree-walking support code */
/*<
      In previous compilers, I had to write custom code for every tree-based
    optimisation, in order to walk down the tree to the right place to find
    the leaves to be optimised.  In this one, I have a generic tree-walking
    procedure which can walk the entire program, but it can be customised 
    so that it takes action only on specific phrases.  This is possible in
    this design only because each set of subphrases stores the count of
    subphrases befoe it - thus allowing a generic tree-walking procedure
    that doesn't have to know what each node consists of until it happens
    across a node of the type it is looking for.

    However this only walks the concrete syntax tree - looks like we need
    something very similar to walk the abstract syntax tree as well...
 */
void walk_analysis_record(int ap, int depth, int wanted(int phraseno), void perform(int ap, int before, int depth))
{
  int i;

  if (wanted(A[ap])) perform(ap, TRUE, depth);
  for (i = 3; i < A[ap+2]+3; i++) {
    if (A[A[ap+i]] >= 512+MAX_BIP) walk_analysis_record(A[ap+i], depth+1, wanted, perform);
  }
  if (wanted(A[ap])) perform(ap, FALSE, depth);
}

int want_all(int phraseno) {
  return TRUE;
}

void print_all(int ap, int before, int depth) {
#ifdef DEBUG_PARSER
  int saved_ap = ap;
  int phrase = A[ap++];
  int alt = A[ap++];
  int phrases = A[ap++]; // defined subphrases
  int i;

  indent(depth, stderr);
  fprintf(stderr, "<%s%s/%d>  ", (before ? "" : "/"), phrasename[phrase-512], alt);
  for (i = 0; i < (3+phrases); i++) {
    fprintf(stderr, "A[%d] = %d, ", saved_ap+i, A[saved_ap+i]);
  }
  fprintf(stderr, "\n");
#endif
}



void walk_AST(int ap, int depth, void perform(int ap, int depth))
{
  int i;

  
  if (opsym(ap) == SEQUENCE) {
      walk_AST(leftchild(ap), depth+1, perform);
      walk_AST(rightchild(ap), depth+1, perform);
  } else perform(ap, depth);

}

void print_trees(int ap, int depth) {
#ifdef DEBUG_PARSER
  if (opsym(ap) != SEQUENCE) {
      draw_tree(ap);
      return;
  }
  if (opsym(leftchild(ap)) != SEQUENCE) {
      draw_tree(leftchild(ap));
  }
  if (opsym(rightchild(ap)) != SEQUENCE) {
      draw_tree(rightchild(ap));
  }
#endif
}

/*>*//*>*/

TRIP compile(int ap, int depth)
{
  /*< Main code-generation procedure.  This is called after parsing, 
     with parameters which describe the parse tree.  By jumping to the
     corresponding statement in the large 'case' below, we execute the
     actions associated with the parse-tree nodes.

     The grammar which was used to build the parse tables is extracted
     from the source below (from comments marked with "//\\") and the
     tables are built with the associated 'takeon' program. (See this
     same directory for the source.  It's quite short...) */

  int saved_ap;
  int phrase;  // A[ap] is the phrase number. A[ap+1] is the alt.
  int alt;     // For consistency, in BIPs, the Alt should always be 1
               // although that may not be the case at the moment :-(
  int phrases; // defined subphrases
  int i;

  TRIP t1, t2, t3, t4, t5; // Temporaries

  // The following ecce command executed on this file will generate varcalc.g:
  // ecce -c "(v.//\\.s..(v/ /e)?m,k)0;%c" varcalc.c varcalc.g
  // May later tweak takeon.c to read from varcalc.c rather than varcalc.g
  // thus simplifying build process and becoming more like yacc.

  saved_ap = ap;
  phrase = A[ap++];
  alt = A[ap++];
  phrases = A[ap++];

  switch (phrase) {

    /*< Built-in phrases */
//\\ # BIPS (Built-in Phrases) are linked to the type-code returned
//\\ # by the line-reconstruction code (aka lexer)
//\\
//\\ # These *must* come first.

// See TYPE_* in first page for the values to use.

//\\
//\\ B<IDENT>=1;
  case P_IDENT:
    return getvar(c[A[ap]].s);

//\\ B<NUM>=5;
  case P_NUM:
    return make_int_const(INT, c[A[ap]].s);

//\\ B<CHARLIT>=3;
  case P_CHARLIT:
    return make_binary_tuple(CONST, INT, (int)*c[A[ap]].s);

/*>*/

//\\
//\\ # Phrase definitions.  SSLIST is the main entry point.
//\\

  case P_SSLIST:
//\\
//\\ P<SSLIST> = <STATEMENT> <SSLIST>,
//\\             ;
    /*< (Click here to expand the code) */
    if (alt == 1) return -1;
    t1 = compile(A[ap], depth+1);
    t2 = compile(A[ap+1], depth+1);
    if (t2 == -1) return t1;
    return make_binary_tuple(SEQUENCE, t1, t2); // break;
    /*>*/

  case P_STATEMENT:
//\\
//\\ P<STATEMENT> = <SIMPLE> <OPTSEMI>;
    return compile(A[ap], depth+1);

  case P_SIMPLE:
//\\
//\\ P<SIMPLE> = <SS>,
//\\             <CONSTDECL>,
//\\             <VARDECL>,
//\\             <DEFFN>;
     return compile(A[ap], depth+1);

  case P_CONSTDECL:
//\\
//\\ P<CONSTDECL> = "const" <IDENT> '=' <NUM>;
return make_binary_tuple(NOOP,  -1, -1);

  case P_VARDECL:
//\\
//\\ P<VARDECL> = "word" <IDENT>;
return make_binary_tuple(NOOP,  -1, -1);

  case P_DEFFN:
//\\
//\\ P<DEFFN> = "let" <IDENT> '(' <FORMALS> ')' '=' <BOOLEXPR>;
    t1 = compile(A[ap], depth+1);
    t2 = compile(A[ap+1], depth+1);
    t3 = compile(A[ap+2], depth+1);
    return make_nary_tuple(DEFFN,  t1, t2, t3);

  case P_FORMALS:
//\\
//\\ P<FORMALS> = <IDENT> <RESTOFFORMALS>;
    t1 = compile(A[ap], depth+1);
    t2 = compile(A[ap+1], depth+1);
    return make_binary_tuple(DEFPARAM, t1, t2);

  case P_RESTOFFORMALS:
//\\
//\\ P<RESTOFFORMALS> = ',' <IDENT> <RESTOFFORMALS>, ;
    if (alt == 1) return make_binary_tuple(NOOP,  -1, -1);
    t1 = compile(A[ap], depth+1);
    t2 = compile(A[ap+1], depth+1);
    return make_binary_tuple(DEFPARAM, t1, t2);

  case P_SS:
//\\
//\\ P<SS> = <IDENT> '=' <BOOLEXPR>;
    t1 = compile(A[ap], depth+1);
    t2 = compile(A[ap+1], depth+1);
    return make_binary_tuple(ASSIGN, t1, t2);

  case P_BOOLEXPR:
//\\
//\\ P<BOOLEXPR> = <BOOLTERM> <RESTOFBOOLTERM>;
    /*< (Click here to expand the code) */

    // 'restofboolterm' will be a binary node of which the left branch is empty -
    // we need to plug that branch with 'boolterm'.

    t1 = compile(A[ap], depth+1);
    t2 = compile(A[ap+1], depth+1);

    if (t2 == -1) return t1; // no restof...

    leftchild(t2) = t1;

    return t2;
    /*>*/

  case P_RESTOFBOOLTERM:
//\\
//\\ P<RESTOFBOOLTERM> = <OROP> <BOOLTERM> <RESTOFBOOLTERM>,
//\\                     ;

    /*< (Click here to expand the code) */
    if (alt == 1) return -1;
    t1 = compile(A[ap], depth+1);
    t2 = compile(A[ap+1], depth+1);
    t3 = compile(A[ap+2], depth+1);

    if (t3 == -1) {
	return make_binary_tuple(leftchild(t1), -1, t2);  // leave a hole for the left op from the parent
    }

    // 'restofboolterm' will be a binary node of which the left branch is empty -
    // we need to plug that branch with 'boolterm'.

    leftchild(t3) = t2;
    return make_binary_tuple(leftchild(t1), -1, t3);
    /*>*/

  case P_BOOLTERM:
//\\
//\\ P<BOOLTERM> = <BOOLFACTOR> <RESTOFBOOLFACTOR>;
    /*< (Click here to expand the code) */
    t1 = compile(A[ap], depth+1);
    t2 = compile(A[ap+1], depth+1);
    if (t2 == -1) return t1; // no restof...

    leftchild(t2) = t1;
    return t2;
    /*>*/

  case P_RESTOFBOOLFACTOR:
//\\
//\\ P<RESTOFBOOLFACTOR> = <ANDOP> <BOOLFACTOR> <RESTOFBOOLFACTOR>,
//\\                       ;

    /*< (Click here to expand the code) */
    if (alt == 1) return -1;
    t1 = compile(A[ap], depth+1);
    t2 = compile(A[ap+1], depth+1);
    t3 = compile(A[ap+2], depth+1);

    if (t3 == -1) return make_binary_tuple(leftchild(t1), -1, t2); // no restof...  Parent will plug in left-hand operator.

    leftchild(t3) = t2;
    return make_binary_tuple(leftchild(t1), -1, t3);
    /*>*/

  case P_BOOLFACTOR:
//\\
//\\ P<BOOLFACTOR> = <OPTNOT> <RELATION>;
    /*< (Click here to expand the code) */
    t1 = compile(A[ap], depth+1);
    t2 = compile(A[ap+1], depth+1);

    if (t1 == -1) return t2;

    return make_unary_tuple(leftchild(t1), t2);
    /*>*/

  case P_OPTNOT:
//\\
//\\ P<OPTNOT> = '!',
//\\             ;
    if (alt == 0) return mkop(NOT);
    return -1;

  case P_OPTSEMI:
//\\
//\\ P<OPTSEMI> = ';',
//\\              ;
    return -1;

  case P_RELATION:
//\\
//\\ P<RELATION> = <EXPR> <RESTOFRELATION>;
    /*< (Click here to expand the code) */
    t1 = compile(A[ap], depth+1);
    t2 = compile(A[ap+1], depth+1);
    if (t2 == -1) return t1; // no restof...

    leftchild(t2) = t1;
    return t2;
    /*>*/

  case P_RESTOFRELATION:
//\\
//\\ P<RESTOFRELATION> = <RELOP> <EXPR>,
//\\                     ;
    /*< (Click here to expand the code) */
    // implicitly "<expr> != 0", if no relop given.
    if (alt == 1) return -1;
    t1 = compile(A[ap], depth+1);
    t2 = compile(A[ap+1], depth+1);
    return make_binary_tuple(leftchild(t1), -1, t2);
    /*>*/

  case P_EXPR:
//\\
//\\ P<EXPR> = <SUM> <RESTOFEXPR>;
    /*< (Click here to expand the code) */
    t1 = compile(A[ap], depth+1);
    t2 = compile(A[ap+1], depth+1);
    if (t2 == -1) return t1;

    leftchild(t2) = t1;
    return t2;
    /*>*/

  case P_RESTOFEXPR:
//\\
//\\ P<RESTOFEXPR> = <SHIFTOP> <SUM> <RESTOFEXPR>,
//\\                 ;
    /*< (Click here to expand the code) */
    if (alt == 1) return -1;
    t1 = compile(A[ap], depth+1);
    t2 = compile(A[ap+1], depth+1);
    t3 = compile(A[ap+2], depth+1);

    if (t3 == -1) return make_binary_tuple(leftchild(t1), -1, t2); // no restof...  Parent will plug in left-hand operator.

    leftchild(t3) = t2;
    return make_binary_tuple(leftchild(t1), -1, t3);
    /*>*/

  case P_SUM:
//\\
//\\ P<SUM> = <OPTADDOP> <TERM> <RESTOFSUM>;
    /*< (Click here to expand the code) */
                                                            // CHECK THIS ONE!  Is it a unary addop????
    t1 = compile(A[ap], depth+1);
    t2 = compile(A[ap+1], depth+1);

    if (t1 != -1) t2 = make_unary_tuple(leftchild(t1), t2);

    t3 = compile(A[ap+2], depth+1);

    if (t3 == -1) return t2; // no restof...

    leftchild(t3) = t2;
    return t3;
    /*>*/

  case P_OPTADDOP:
//\\
//\\ P<OPTADDOP> = <ADDOP>,
//\\               ;
    /*< (Click here to expand the code) */
    if (alt == 1) return -1;
    return compile(A[ap], depth+1);
    /*>*/

  case P_RESTOFSUM:
//\\
//\\ P<RESTOFSUM> = <ADDOP> <TERM> <RESTOFSUM>,
//\\                ;
    /*< (Click here to expand the code) */
    if (alt == 1) return -1;
    t1 = compile(A[ap], depth+1);
    t2 = compile(A[ap+1], depth+1);
    t3 = compile(A[ap+2], depth+1);

    if (t3 == -1) return make_binary_tuple(leftchild(t1), -1, t2); // no restof...  Parent will plug in left-hand operator.

    leftchild(t3) = t2;
    return make_binary_tuple(leftchild(t1), -1, t3);
    /*>*/

  case P_TERM:
//\\
//\\ P<TERM> = <FACTOR> <RESTOFTERM>;
    /*< (Click here to expand the code) */
    t1 = compile(A[ap], depth+1);
    t2 = compile(A[ap+1], depth+1);
    if (t2 == -1) return t1; // no restof...

    leftchild(t2) = t1;
    return t2;
    /*>*/

  case P_RESTOFTERM:
//\\
//\\ P<RESTOFTERM> = <MULOP> <FACTOR> <RESTOFTERM>,
//\\                 ;
    /*< (Click here to expand the code) */
    if (alt == 1) return -1;
    t1 = compile(A[ap], depth+1);
    t2 = compile(A[ap+1], depth+1);
    t3 = compile(A[ap+2], depth+1);

    if (t3 == -1) return make_binary_tuple(leftchild(t1), -1, t2); // no restof...  Parent will plug in left-hand operator.

    leftchild(t3) = t2;
    return make_binary_tuple(leftchild(t1), -1, t3);
    /*>*/

  case P_OROP:
//\\
//\\ P<OROP> = '||';
     return mkop(OR);

  case P_ANDOP:
//\\
//\\ P<ANDOP> = '&&';
     return mkop(AND);

  case P_RELOP:
//\\
//\\ P<RELOP> = '<>', '<=', '<', '>=', '>', '!=', '=';
    switch (alt) {
        case 0: return mkop(NE); // elide?
	case 5: return mkop(NE);
	case 1: return mkop(LE);
	case 2: return mkop(LT);
	case 3: return mkop(GE);
	case 4: return mkop(GT);
	case 6: return mkop(EQ);
    }


  case P_SHIFTOP:
//\\
//\\ P<SHIFTOP> = '<<', '>>';
    return (alt == 0 ? mkop(LSH) : mkop(RSH));

  case P_ADDOP:
//\\
//\\ P<ADDOP> = '+', '-';
    return (alt == 0 ? mkop(ADD) : mkop(SUB));

  case P_MULOP:
//\\
//\\ P<MULOP> = '*', '/';
    return (alt == 0 ? mkop(MUL) : mkop(DIV));

  case P_FACTOR:
//\\
//\\ P<FACTOR> = '(' <BOOLEXPR> ')',
//\\             <IDENT> '(' <PARAMLIST> ')',
//\\             <IDENT>,
//\\             <NUMBER>;
    if (alt == 0) {
      return compile(A[ap], depth+1);
    } else if (alt == 1) {
      t1 = compile(A[ap], depth+1); // VAR (ident is a var; leftchild of ident is a tag; rightchild of tag is index into stringpool)
      t2 = compile(A[ap+1], depth+1); // expr param
      t3 = leftchild(t1); // TAG (fn name)
      t4 = rightchild(t3); // stringpool offset
      return make_binary_tuple(FNCALL, make_proc_name(t4), t2);
    } else if (alt == 2) {
      return compile(A[ap], depth+1);
    } else { // alt = 3
      return compile(A[ap], depth+1);
    }


  case P_PARAMLIST:
//\\
//\\  P<PARAMLIST> = <EXPR> <RESTOFPARAMLIST>,
//\\               ;
//\\

    if (alt == 1) return make_binary_tuple(NOOP,  -1, -1);
    t1 = compile(A[ap], depth+1);
    t2 = compile(A[ap+1], depth+1);
    return make_binary_tuple(PARAM, t1, t2);    


  case P_RESTOFPARAMLIST:
//\\
//\\  P<RESTOFPARAMLIST> = ',' <EXPR> <RESTOFPARAMLIST>,
//\\                ;
//\\

    if (alt == 1) return make_binary_tuple(NOOP,  -1, -1);
    t1 = compile(A[ap], depth+1);
    t2 = compile(A[ap+1], depth+1);
    return make_binary_tuple(PARAM, t1, t2);    
    

  case P_NUMBER:
//\\
//\\ P<NUMBER> = <NUM>,
//\\             <CHARLIT>;
    return compile(A[ap], depth+1);

//\\
//\\ E
//\\ # 'E' is end of grammar.  Everything after this is ignored.
  default:
      fprintf(stderr, "*** Internal error at line %d.  ap=%d  phrase=%d", __LINE__, ap, phrase);
      if (((phrase-512) >= 0) && ((phrase-512) < MAX_PHRASE)) fprintf(stderr, "\n    (possible missing \"case P_%s:\" in compile()?)", phrasename[phrase-512]);
      fprintf(stderr, "\n");
      t5=t5/0;
      exit(2);

  }

  return(-1); // DUMMY TRIP, NOTHING TO RETURN
  /*>*/
}
/*>*//*>*/

/*< Code generators - one for three-address, one for a stack machine, and one run-time interpreter */
static int nextlab = 0;

/*< basic codegen_three_address() generates a simple 3-address intermediate code */
// Codegen is the guts of the compiler, which effectively serialises
// the AST into sequentially executable statements.  A further phase
// is required to actually generate executable code.

/*< Support I/O procs for codegen */
void declare(char *name)
{
    fprintf(stdout, "    int %s;\n", name);
}

void loadconst(char *format, TRIP dest, int value)
{
    fprintf(stdout, "    ");
    fprintf(stdout, format, dest, value);
}

void loadvar(char *format, TRIP dest, char *varname)
{
    fprintf(stdout, "    ");
    fprintf(stdout, format, dest, varname);
}
void store(char *format, char *varname, TRIP source)
{
    fprintf(stdout, "    ");
    fprintf(stdout, format, varname, source);
}
void operate(char *format, TRIP dest, TRIP leftop, char *op, TRIP rightop)
{
    fprintf(stdout, "    ");
    fprintf(stdout, format, dest, leftop, op, rightop);
}
void monoperate(char *format, TRIP dest, char *op, TRIP leftop)
{
    fprintf(stdout, "    ");
    fprintf(stdout, format, dest, op, leftop);
}
void put_goto(int lab)
{
    fprintf(stdout, "    ");
    fprintf(stdout, "goto L%02d;\n", lab);
}
void put_ifgoto(int cond, int lab, int sense)
{
    fprintf(stdout, "    ");
    fprintf(stdout, "if (%s_t%d) goto L%02d;\n", (sense ? "" : "!"), cond, lab);
}
void input(TRIP i)
{                              /* TACC BUG!  CANNOT ESCAPE double quotes properly */
    fprintf(stdout, "    ");
    fprintf(stdout, "fprintf(stderr, %c%s: %c); fflush(stderr); fscanf(stdin, %c%%d%c, &%s);\n",
	    '"', nameof(leftchild(i)), '"',
	    '"', '"', nameof(leftchild(i)));
}
void print(TRIP i)
{
    fprintf(stdout, "    ");
    fprintf(stdout, "fprintf(stdout, %c%%d\\n%c, _t%d); fflush(stdout);\n", '"', '"', leftchild(i));
}
// this macro was used in debugging, not really needed now.
#define put_label(n) xput_label(n, __LINE__)
void xput_label(int lab, int line)
{
    fprintf(stdout, "L%02d:\n", lab);
}
/*>*/

void codegen_three_address(TRIP root) {

    if (root == -1) return;
    switch (opsym(root)) {

	case NOOP:
	    break;

	case CONST:
	    loadconst("_t%d = %d;\n", root, rightchild(root));
	    break;

	case DECLARE:
	    root = leftchild(root);
	    declare(stringpool+rightchild(leftchild(root)));
	    break;

	case VAR:
	    loadvar("_t%d = %s;\n", root, nameof(root));
	    break;

	case ASSIGN:
	    codegen_three_address(rightchild(root));
	    store("%s = _t%d;\n", nameof(leftchild(root)), rightchild(root));
	    break;

	case PARAM:
	    codegen_three_address(leftchild(root));
	    codegen_three_address(rightchild(root));
	    break;

	case SEQUENCE:
	    codegen_three_address(leftchild(root));
	    codegen_three_address(rightchild(root));
	    break;

	case INPUT:
	    input(root);
	    break;

	case PRINT:
	    codegen_three_address(leftchild(root));
	    print(root);
	    break;

	case NEG:
	case NOT:
	    codegen_three_address(leftchild(root));
	    monoperate("_t%d = %s_t%d;\n", root, c_infix_op[opsym(root)], leftchild(root));
	    break;

	default:
	    /* Be careful not to default anything other than binary operators! */
	    if (arity[opsym(root)] != 3) {
		fprintf(stdout, "*** Not Implemented: codegen_three_address(%s)\n", name[opsym(root)]);
		break;
	    }

	    codegen_three_address(leftchild(root));
	    codegen_three_address(rightchild(root));
	    operate("_t%d = (_t%d %s _t%d);\n", root,
		    leftchild(root), c_infix_op[opsym(root)], rightchild(root));
	    break;
    }
}
/*>*/

/*< codegen_stack generates for a stack machine */
void codegen_stack(TRIP root) {
static char *current_function_name=""; // temp hack.  real soln involves a scope/block-stack
static int code_started = FALSE;

    if ((root == -1) || (opsym(root) == NOOP)) return;

    if ((opsym(root) == ASSIGN) && !code_started) { // add all top-level opcodes
       // first actual code (we skip over function definitions)
       fprintf(stdout, "\n");
       fprintf(stdout, "_start:                    ; Main program entry\n"); code_started = TRUE;
    }

    switch (opsym(root)) {

	case CONST:
	    fprintf(stdout, "          PUSH #%d\n", rightchild(root));
	    break;

	case VAR:
	    fprintf(stdout, "          PUSH %s\n", nameof(root));
	    break;

	case ASSIGN:
	    codegen_stack(rightchild(root));
	    fprintf(stdout, "          POP %s\n", nameof(leftchild(root)));
	    break;

      case DEFFN:
          current_function_name = stringpool+rightchild(leftchild(leftchild(root)));
          fprintf(stdout, "%s:\n", current_function_name);
          // first we use the paramlist in rightchild to push a temporary definition
          // of each parameter on the namespace stack

          // push_temporary_parameter_definitions(rightchild(root)); // (and also generate code to pop params from stack to locals)
	    codegen_stack(rightchild(root));

          // then we compile code for the definition, which will pick up local parameters
          // for those idents rather than any globals of the same name
          // first pop the params off the data stack and assign to locals.
          // later implementation won't use static locals but will pick up off stack directly

	    codegen_stack(nthchild(root, 3));

          // restore_temporary_parameter_definitions(rightchild(root));

	    fprintf(stdout, "          RET\n");  // NOTE WE'RE ONLY HANDLING ONE IMPLICIT RESULT BY DEFAULT
          break;

      case DEFPARAM:
          // small tweak to ensure parameters popped in reverse order to undo pushes to stack
	    codegen_stack(rightchild(root));
	    if (opsym(leftchild(root)) == DEFPARAM) codegen_stack(leftchild(root)); else {
            int formal = rightchild(leftchild(leftchild(root)));
            int local;
            static char locals[128];
            sprintf(locals, "%s_%s", current_function_name, stringpool+formal);

            strcpy(stringpool+nextstring, locals); /* Create a backstop for when not found */
            local = str_to_pool(locals);
            if (local != nextstring) {
	        fprintf(stderr, "ERROR: local parameter '%s' already exists.\n", locals);
            } else nextstring += strlen(locals)+1; /* Not found, add it */

            rightchild(leftchild(leftchild(root))) = local;  // replace for the context of this fn def
            
            // the parameter at leftchild(root) is a "VAR" - currently we can cheat
            // and dive into the string directly, because out model of params (and variables)
            // is so dumb.  However when we start using types properly this will have to change. 
            fprintf(stdout, "          POP %s\n", locals);

          }
	    break;

	case PARAM:
	    codegen_stack(leftchild(root));
	    codegen_stack(rightchild(root));
	    break;

      case SEQUENCE:
	    codegen_stack(leftchild(root));
	    codegen_stack(rightchild(root));
	    break;

      case FNCALL:
	    codegen_stack(rightchild(root));  // zero or more params
	    fprintf(stdout, "          CALL %s\n", stringpool+rightchild(leftchild(root)));
	    break;

	case NEG:
	case NOT:
	    codegen_stack(leftchild(root));
	    fprintf(stdout, "          %s\n", name[opsym(root)]);
	    break;

	default:
	    /* Be careful not to default anything other than binary operators! */
	    if (arity[opsym(root)] != 3) {
		fprintf(stdout, "*** Not Implemented: codegen_stack(%s)\n", name[opsym(root)]);
		break;
	    }

	    codegen_stack(leftchild(root));
	    codegen_stack(rightchild(root));
	    fprintf(stdout, "          %s\n", name[opsym(root)]);
	    break;
    }
}
/*>*/

/*< Run-time interpreter: Cheap & nasty code execution directly from the AST! */

// Trivial run-time stack implementation:
static int stack[128];
static int stackp = -1;
void Push(int val) {
    stack[++stackp] = val;
}
int Pop(void) {
    return stack[stackp--];
}


/*< (a little hack) */
// What I should be doing here is making a VAR tuple one unit larger, with the
// extra unit pointing to the memory address/offset where the var is stored.
// This was in fact always part of the plan (along with more complex type information)
// but at this stage the code is really just a placeholder.
// Meanwhile, by having an array the same size as the AST, we have an obvious
// place where we can store the data, and for now do not need to implement
// the DECLARE opcode in the interpreter below...

int wasteful_array[MAXTRIPS];
#define variable_contents(memloc) wasteful_array[memloc]
/*>*/

void execute_AST(TRIP root) {
    TRIP left, right;

    if (root == -1) return;
    switch (opsym(root)) {

	case NOOP:
	    break;

	case CONST:
	    Push(rightchild(root));
	    break;

	case VAR:
	    Push(variable_contents(root));
	    break;

	case ASSIGN:

// Until I add a "print" command to this language, we can see the results of
// computations by a simple hack, which is to print out the value of any assignments.

	    execute_AST(rightchild(root));
	    variable_contents(leftchild(root)) = Pop();

	    fprintf(stdout, "   %s = %d\n",
                            nameof(leftchild(root)),
                            variable_contents(leftchild(root)));

	    break;

	case SEQUENCE:
	    execute_AST(leftchild(root));
	    execute_AST(rightchild(root));
	    break;

	case NEG:
	    execute_AST(leftchild(root));
            Push(-Pop());
	    break;

	case NOT:
	    execute_AST(leftchild(root));
            Push(!Pop()); // Boolean NOT, not bitwise NOT
	    break;

//    AND, OR,

        case AND:
	    execute_AST(leftchild(root));
	    execute_AST(rightchild(root));
	    right = Pop(); left = Pop(); Push(left && right);
	    break;

        case OR:
	    execute_AST(leftchild(root));
	    execute_AST(rightchild(root));
	    right = Pop(); left = Pop(); Push(left || right);
	    break;

//    ADD, SUB, MUL, DIV,
        case ADD:
	    execute_AST(leftchild(root));
	    execute_AST(rightchild(root));
	    right = Pop(); left = Pop(); Push(left + right);
	    break;

        case SUB:
	    execute_AST(leftchild(root));
	    execute_AST(rightchild(root));
	    right = Pop(); left = Pop(); Push(left - right);
	    break;

        case MUL:
	    execute_AST(leftchild(root));
	    execute_AST(rightchild(root));
	    right = Pop(); left = Pop(); Push(left * right);
	    break;

        case DIV:
	    execute_AST(leftchild(root));
	    execute_AST(rightchild(root));
	    right = Pop(); left = Pop();
            if (right == 0) {
		fprintf(stderr, "Run-time error: divide by zero\n"); exit(1);
	    }
            Push((int)(left / right));
	    break;

//    LSH, RSH, EXP,
        case LSH:
	    execute_AST(leftchild(root));
	    execute_AST(rightchild(root));
	    right = Pop(); left = Pop(); Push(left << right);
	    break;

        case RSH:
	    execute_AST(leftchild(root));
	    execute_AST(rightchild(root));
	    right = Pop(); left = Pop(); Push(left >> right);
	    break;

        case EXP:
	    execute_AST(leftchild(root));
	    execute_AST(rightchild(root));
	    right = Pop(); left = Pop();
            //Push(iexp(left, right));
            {int temp = left;
		while (right >= 2) {
                    left = left * temp; right -= 1;
		}
	    }
	    Push(left);
            break;

//    EQ, NE, LT, GT, LE, GE,
        case EQ:
	    execute_AST(leftchild(root));
	    execute_AST(rightchild(root));
	    right = Pop(); left = Pop(); Push(left == right);
	    break;

        case NE:
	    execute_AST(leftchild(root));
	    execute_AST(rightchild(root));
	    right = Pop(); left = Pop(); Push(left != right);
	    break;

        case LT:
	    execute_AST(leftchild(root));
	    execute_AST(rightchild(root));
	    right = Pop(); left = Pop(); Push(left < right);
	    break;

        case GT:
	    execute_AST(leftchild(root));
	    execute_AST(rightchild(root));
	    right = Pop(); left = Pop(); Push(left > right);
	    break;

	case LE:
	    execute_AST(leftchild(root));
	    execute_AST(rightchild(root));
	    right = Pop(); left = Pop(); Push(left <= right);
	    break;

        case GE:
	    execute_AST(leftchild(root));
	    execute_AST(rightchild(root));
	    right = Pop(); left = Pop(); Push(left >= right);
	    break;

	default:
            // INTERNAL ERROR!
	    fprintf(stdout, "*** Not Implemented: execute_AST(%s)\n", name[opsym(root)]);
            exit(1);
	    break;

    }
    //fprintf(stdout, "[%d] ", stack[stackp]);
    //printtrip(root); // look at the op we just executed
}
/*>*/
/*>*/
/*>*/

int main(int argc, char **argv) {
char *s;

  /*< Handle program arguments */

  /*< Get clean version of executable name.  Should work on most existing systems (2006) */
  progname = strdup(argv[0]);
  if ((s = strrchr(progname, '/')) != NULL) progname = s+1;  // Unix
  if ((s = strrchr(progname, '\\')) != NULL) progname = s+1; // M$
  if ((s = strrchr(progname, ']')) != NULL) progname = s+1;  // Dec
  if ((s = strrchr(progname, ';')) != NULL) *s = '\0';       // Version no's
  if (((s = strrchr(progname, '.')) != NULL) && (strcasecmp(s, ".exe") == 0)) *s = '\0';
  if (((s = strrchr(progname, '.')) != NULL) && (strcasecmp(s, ".com") == 0)) *s = '\0';
/*>*/

  if ((argc == 3) && strcmp(argv[1], "-d") == 0) {
    argv++; argc--; debug_parser = TRUE;
  }

  if (argc != 2) {
    fprintf(stderr, "syntax: %s [-d] filename\n", progname);
    exit(1);
  }

  sourcefile = fopen(argv[1], "r");
  if (sourcefile == NULL) {
    fprintf(stderr, "%s: %s - %s\n", progname, strerror(errno), argv[1]);
    exit(errno);
  }

  curfile = argv[1]; startline = TRUE; whitespace = TRUE;
  onecharstr = (char *)malloc(512);
/*>*/

  /*< Lexical scan */
  line_reconstruction(); // Effectively, lexing.

  /*< Debug the lexed tokens? */
#ifdef DEBUG_PARSER
  if (debug_parser) {
    int i; // DEBUG ONLY
    fprintf(stderr, "\nLexical token stream:\n\n");
    for (i = 0; i < nextfree; i++) {
      fprintf(stderr, "C[%d] => %s, line %d, col %d: [%0d] %s\n",
                          i,  c[i].f, c[i].l, c[i].col, c[i].t, c[i].s);
    }
  }
#endif
/*>*//*>*/

  /*< Call the parser */

#ifdef DEBUG_TRIPS_DURING
  fprintf(stderr, "Trips before patching holes:\n");
#endif

  if (!parse(PHRASE_BASE, 0)) {
    /*< Attempt to print a sensible error if the parse failed */
    if (bestparse == nextfree) {
      fprintf(stderr, "\"%s\", Line %d, Col %d: Premature end of file while looking for %s\n",
                       argv[1], c[bestparse].l, c[bestparse].col+1, looking_for);
    } else {
      int i;
      fprintf(stderr, "\"%s\", Line %d, Col %d: Syntax error while looking for %s near ",
                       argv[1], c[bestparse].l, c[bestparse].col+1, looking_for);
      for (i = bestparse; i < bestparse+3; i++) {
        if (i == nextfree) {
          fprintf(stderr, "<End of file>");
          break;
        }
        switch (c[i].t) {
        case TYPE_HEXINT:
          fprintf(stderr, "$"); // *OR* ... We could put the '$' back in front of the string
          /* drop through */    // and probably save much code whenever printing.  Use str+1
        case TYPE_TAG:
        case TYPE_CHAR:
        case TYPE_INT:
        case TYPE_KEYWORD:
          fprintf(stderr, "%s", c[i].s);
          break;
        case TYPE_STRING:
          fprintf(stderr, "\"%s\"", c[i].s);
          break;
        case TYPE_CHARCONST:
          fprintf(stderr, "'%s'", c[i].s);
          break;
        }
        fprintf(stderr, (i == (bestparse+2) ? " ..." : " "));
      }
      fprintf(stderr, "\n");
    }
    /*>*/
    exit(1);
  }/*>*/

  /*< Generate code */

  /*< Debugging */
#ifdef DEBUG_PARSER
  if (debug_parser) walk_analysis_record(0, 0, want_all, print_all); // Diags: print final parse tree
#endif
/*>*/

  {int program;

     /*< Debugging */

#ifdef DEBUG_TRIPS_DURING
     fprintf(stdout, "Trips as they are created (before patching holes):\n"); fflush(stdout);
#endif
/*>*/

     program = compile(0, 0);

     /*< Debugging */

#ifdef DEBUG_TRIPS_AFTER
     fprintf(stdout, "\n\nAST (with trips after patching holes):\n\n"); fflush(stdout);
     {int i, lasttrip; for (i = 0; i < nexttrip; i += arity[AST[i]]) printtrip(lasttrip = i);
     fprintf(stdout, "\n\n"); fflush(stdout);
#ifdef DEBUG_TREES
     walk_AST(lasttrip, 0, print_trees);
#endif
     }
#endif
/*>*/

     // Now generate the output code from the AST.

//     fprintf(stdout, "\nAST serialised into three-address code:\n\n");
//     codegen_three_address(program);

     //fprintf(stdout, "\nStack-based code:\n\n"); fflush(stdout);
     {
	 int i, l;
         l = c[0].l;
         for (i = 0; i < nextfree; i++) {
	     if (c[i].l != l) {fprintf(stdout, "\n"); l = c[i].l;}
	     fprintf(stdout, "%s ", c[i].s);
         }
         fprintf(stdout, "\n\n");
     }
     codegen_stack(program);

//     fprintf(stdout, "\nRun-time interpretation:\n\n");
//     execute_AST(program);
     fprintf(stdout, "\n\n");

  }
  /*>*/

  exit(0); return(1);
}
